#ifndef FSGUIBITMAP_IS_INCLUDED
#define FSGUIBITMAP_IS_INCLUDED
/* { */

#include "ysbitmap.h"

/*! This class makes an array of bitmap.  Can be used for breaking down a large bitmap into 
    pieces of bitmaps that each can be drawn by a poor-man's graphics toolkit such as OpenGL ES.
    Typical OpenGL ES cannot draw any bitmap greater than 64x64, and the size need to be 2^n times 2^m.

	Note that the tile in the last row or last column may not have a same size as the other tiles.
	For example, totalWidth is 260 and the maxWidth is 16, the width of the last tile will be 4.
	Size will always be 2^n.
    */
class YsBitmapArray
{
private:
	unsigned int totalWid,totalHei;
	unsigned int tileWid,tileHei;
	unsigned int nDivX,nDivY;
	YsBitmap *bmpArray;

public:
	/*! Default constructor. */
	YsBitmapArray();

	/*! Copy constructor .*/
	YsBitmapArray(const YsBitmapArray &from);

	/*! Copy operator. */
	const YsBitmapArray &operator=(const YsBitmapArray &from);

private:
	void CopyFrom(const YsBitmapArray &from);

public:
	/*! Default destructor. */
	~YsBitmapArray();

	/*! Initializes the bitmap array. */
	void CleanUp(void);

	/*! Copys a bitmap to this bitmap array. The bitmap will be divided up into an array of bitmaps.
	    One piece of a bitmap will be 2^n times 2^m pixels that does not exceed maxWidth, maxHeight. */
	YSRESULT SetAndBreakUp(const YsBitmap &from,int maxWidth,int maxHeight);

private:
	void ResizeArray(int n);

public:
	/*! Returns the total width of the bitmap. */
	unsigned int GetTotalWidth(void) const;

	/*! Returns the total height of the bitmap. */
	unsigned int GetTotalHeight(void) const;

	/*! Returns number of pieces of bitmap in X direction */
	unsigned int GetNumBlockX(void) const;

	/*! Returns number of pieces of bitmap in Y direction */
	unsigned int GetNumBlockY(void) const;

	/*! Returns width of a tile in number of pixels */
	unsigned int GetTileWidth(void) const;

	/*! Returns height of a tile in number of pixels */
	unsigned int GetTileHeight(void) const;

	/*! Returns a constant pointer to a tile bitmap.  If x or y is not in range,
	    it will return NULL. */
	const YsBitmap *GetTileBitmap(unsigned int x,unsigned int y) const;

	/*! Returns a pointer to a tile bitmap.  If x or y is not in range,
	    it will return NULL. */
	YsBitmap *GetTileBitmap(unsigned int x,unsigned int y);
};

/* } */
#endif
