#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "ysglsl3ddrawing.h"
#include "ysglslutil.h"
#include "ysglmath.h"
#include "ysglbitmap.h"

/* Force Visual C++ to type-mismatching error. */
#pragma warning( error : 4028)
#pragma warning( error : 4047)

#ifdef __APPLE_IOS__
#define YS_GLSL_LOWP "lowp "
#define YS_GLSL_MIDP "mediump "
#define YS_GLSL_HIGHP "highp "
#else
#define YS_GLSL_LOWP
#define YS_GLSL_MIDP
#define YS_GLSL_HIGHP
#endif

#define YSGLSLMAXNUMLIGHT 8
#define YSGLSLMAXNUMLIGHTSTR "8"


struct YsGLSL3DRenderer
{
	int inUse;

	// Shader and Program
	GLuint programId;
	GLuint vertexShaderId,fragmentShaderId;

	// Uniforms
	GLint uniformProjectionPos;
	GLint uniformModelViewPos;
	GLint uniformOffsetTransformPos;

	GLint uniformLightEnabledPos;     // -1 if no lighting   bool lightEnabled[YSGLMAXNUMLIGHT]
	GLint uniformLightPositionPos;    // -1 if no lighting   vec4 lightPos[YSGLMAXNUMLIGHT]
	GLint uniformLightColorPos;       // -1 if no lighting   vec4 lightColor[YSGLMAXNUMLIGHT]
	GLint uniformSpecularColorPos;    // -1 if no lighting   vec3 specular
	GLint uniformSpecularExponentPos; // -1 if no lighting   float specularExponent
	GLint uniformAmbientColorPos;     // 
	GLint uniformAlphaCutOffPos;

	GLint uniformFogEnabledPos;
	GLint uniformFogDensityPos;
	GLint uniformFogColorPos;

	GLint uniformDimensionPos;  // For flash renderer
	GLint uniformRadius1Pos;  // For flash renderer
	GLint uniformRadius2Pos;  // For flash renderer
	GLint uniformMaxIntensityPos;  // For flash renderer
	GLint uniformAngleIntervalPos;  // For flash renderer
	GLint uniformAngleOffsetPos;  // For flash renderer
	GLint uniformExponentPos;  // For flash renderer

	GLint uniformColorPos;       // -1 if VariColor

	// Texture mapping >>
	GLint uniformTextureTypePos;
	GLint uniformUseNegativeOneToPositiveOnTexCoordPos;

	GLint uniformTextureTileTransformPos;
	GLint uniformTexturePos;

	GLint uniformBillBoardClippingTypePos;
	GLint uniformBillBoardCenterPos;
	GLint uniformBillBoardTransformTypePos;
	GLint uniformBillBoardDimensionPos;
	// Texture mapping <<

	GLint uniformUseZOffsetPos;
	GLint uniformZOffsetPos;
	// Memo for adding uniforms:
	//   Uniform position must be set to -1 in YsGLSLCreateAndInitialize3DRenderer

	// Attributes
	GLint attribVertexPos;
	GLint attribColorPos;        // -1 if MonoColor
	GLint attribNormalPos;       // -1 if no normal
	GLint attribTexCoordPos;     // -1 if no tex coord
	GLint attribViewOffsetPos;   // -1 if no view offset (offset vector in the view coordinate)
	GLint attribPixelOffsetPos;  // -1 if no pixel offset
	// Memo for adding attributes:
	//   Attribute position must be set to -1 in YsGLSLCreateAndInitialize3DRenderer
	//   Attributes array needs to be enabled in YsGLSLUse3DRenderer and disabled in YsGLSLEndUse3DRenderer.
	//   When an attribute is added, make sure enabling/disabling are also added in the above functions.

	// Saving transformations
	double projectionMatrix[16];
	double modelViewMatrix[16];

	// Saving lighting
	GLfloat ambientColor[3];
	GLint lightEnabled[YSGLSLMAXNUMLIGHT];
	GLfloat lightPosition[4*YSGLSLMAXNUMLIGHT];
	GLfloat lightColor[3*YSGLSLMAXNUMLIGHT];
	GLfloat specularExponent;
	GLfloat specularColor[3];

	// Saving fog density
	GLfloat fogDensity;
};

static struct YsGLSL3DRenderer *YsGLSLCreateAndInitialize3DRenderer(void)
{
	struct YsGLSL3DRenderer *renderer;
	renderer=(struct YsGLSL3DRenderer *)malloc(sizeof(struct YsGLSL3DRenderer));

	renderer->inUse=0;

	renderer->programId=0;
	renderer->vertexShaderId=0;
	renderer->fragmentShaderId=0;

	renderer->uniformProjectionPos=-1;
	renderer->uniformModelViewPos=-1;
	renderer->uniformOffsetTransformPos=-1;
	renderer->uniformAlphaCutOffPos=-1;

	renderer->uniformLightEnabledPos=-1;
	renderer->uniformLightPositionPos=-1;
	renderer->uniformLightColorPos=-1;
	renderer->uniformSpecularColorPos=-1;
	renderer->uniformSpecularExponentPos=-1;
	renderer->uniformAmbientColorPos=-1;

	renderer->uniformFogEnabledPos=-1;
	renderer->uniformFogDensityPos=-1;
	renderer->uniformFogColorPos=-1;

	renderer->uniformDimensionPos=-1;
	renderer->uniformRadius1Pos=-1;
	renderer->uniformRadius2Pos=-1;
	renderer->uniformMaxIntensityPos=-1;
	renderer->uniformAngleIntervalPos=-1;
	renderer->uniformAngleOffsetPos=-1;
	renderer->uniformExponentPos=-1;

	renderer->uniformColorPos=-1;

	renderer->uniformTextureTypePos=-1;
	renderer->uniformUseNegativeOneToPositiveOnTexCoordPos=-1;

	renderer->uniformTextureTileTransformPos=-1;
	renderer->uniformTexturePos=-1;

	renderer->uniformBillBoardClippingTypePos=-1;
	renderer->uniformBillBoardCenterPos=-1;
	renderer->uniformBillBoardTransformTypePos=-1;
	renderer->uniformBillBoardDimensionPos=-1;


	renderer->uniformUseZOffsetPos=-1;
	renderer->uniformZOffsetPos=-1;


	renderer->attribVertexPos=-1;
	renderer->attribColorPos=-1;
	renderer->attribNormalPos=-1;
	renderer->attribTexCoordPos=-1;
	renderer->attribViewOffsetPos=-1;
	renderer->attribPixelOffsetPos=-1;

	renderer->fogDensity=1.0;

	return renderer;
}

static void YsGLSLGetTextureUniformPos(struct YsGLSL3DRenderer *renderer)
{
	renderer->uniformTextureTypePos=glGetUniformLocation(renderer->programId,"textureType");
	renderer->uniformUseNegativeOneToPositiveOnTexCoordPos=glGetUniformLocation(renderer->programId,"useNegativeOneToPositiveOneTexCoord");

	renderer->uniformTexturePos=glGetUniformLocation(renderer->programId,"textureIdent");

	renderer->uniformTextureTileTransformPos=glGetUniformLocation(renderer->programId,"textureTileTransform");

	renderer->uniformBillBoardCenterPos=glGetUniformLocation(renderer->programId,"billBoardCenter");
	renderer->uniformBillBoardTransformTypePos=glGetUniformLocation(renderer->programId,"billBoardTransformType");
	renderer->uniformBillBoardDimensionPos=glGetUniformLocation(renderer->programId,"billBoardDimension");
}

void YsGLSLDelete3DRenderer(struct YsGLSL3DRenderer *renderer)
{
	if(NULL!=renderer)
	{
		glDeleteProgram(renderer->programId);
		glDeleteShader(renderer->vertexShaderId);
		glDeleteShader(renderer->fragmentShaderId);

		free(renderer);
	}
}

void YsGLSLSet3DRendererProjectionfv(struct YsGLSL3DRenderer *renderer,const GLfloat mat[16])
{
	if(NULL!=renderer)
	{
		int i;
		glUniformMatrix4fv(renderer->uniformProjectionPos,1,GL_FALSE,mat);
		for(i=0; i<16; ++i)
		{
			renderer->projectionMatrix[i]=mat[i];
		}
	}
}

void YsGLSLSet3DRendererProjectiondv(struct YsGLSL3DRenderer *renderer,const double mat[16])
{
	if(NULL!=renderer)
	{
		int i;
		GLfloat matf[16];
		for(i=0; i<16; ++i)
		{
			matf[i]=(GLfloat)mat[i];
			renderer->projectionMatrix[i]=mat[i];
		}
		glUniformMatrix4fv(renderer->uniformProjectionPos,1,GL_FALSE,matf);
	}
}

void YsGLSLGet3DRendererProjectionfv(GLfloat mat[16],const struct YsGLSL3DRenderer *renderer)
{
	if(NULL!=renderer)
	{
		int i;
		for(i=0; i<16; ++i)
		{
			mat[i]=(GLfloat)renderer->projectionMatrix[i];
		}
	}
}

void YsGLSLGet3DRendererProjectiondv(double mat[16],const struct YsGLSL3DRenderer *renderer)
{
	if(NULL!=renderer)
	{
		int i;
		for(i=0; i<16; ++i)
		{
			mat[i]=renderer->projectionMatrix[i];
		}
	}
}

void YsGLSLSet3DRendererModelViewfv(struct YsGLSL3DRenderer *renderer,const GLfloat mat[16])
{
	if(NULL!=renderer)
	{
		int i;
		for(i=0; i<16; ++i)
		{
			renderer->modelViewMatrix[i]=mat[i];
		}
		glUniformMatrix4fv(renderer->uniformModelViewPos,1,GL_FALSE,mat);
	}
}

void YsGLSLSet3DRendererModelViewdv(struct YsGLSL3DRenderer *renderer,const double mat[16])
{
	if(NULL!=renderer)
	{
		int i;
		GLfloat matf[16];
		for(i=0; i<16; ++i)
		{
			matf[i]=(GLfloat)mat[i];
			renderer->modelViewMatrix[i]=mat[i];
		}
		glUniformMatrix4fv(renderer->uniformModelViewPos,1,GL_FALSE,matf);
	}
}

void YsGLSLGet3DRendererModelViewfv(GLfloat mat[16],const struct YsGLSL3DRenderer *renderer)
{
	if(NULL!=renderer)
	{
		int i;
		for(i=0; i<16; ++i)
		{
			mat[i]=(GLfloat)renderer->modelViewMatrix[i];
		}
	}
}

void YsGLSLGet3DRendererModelViewdv(double mat[16],const struct YsGLSL3DRenderer *renderer)
{
	if(NULL!=renderer)
	{
		int i;
		for(i=0; i<16; ++i)
		{
			mat[i]=renderer->modelViewMatrix[i];
		}
	}
}

YSRESULT YsGLSLSet3DRendererUniformColorfv(const struct YsGLSL3DRenderer *renderer,const GLfloat color[4])
{
	if(NULL!=renderer)
	{
		if(0<=renderer->uniformColorPos)
		{
			glUniform4fv(renderer->uniformColorPos,1,color);
			return YSOK;
		}
		else
		{
			return YSERR;
		}
	}
	return YSERR;
}

YSRESULT YsGLSLSet3DRendererAlphaCutOff(const struct YsGLSL3DRenderer *renderer,GLfloat alphaCutOff)
{
	if(NULL!=renderer && 0<=renderer->uniformAlphaCutOffPos)
	{
		glUniform1f(renderer->uniformAlphaCutOffPos,alphaCutOff);
		return YSOK;
	}
	return YSERR;
}

YSRESULT YsGLSLSet3DRendererTextureType(const struct YsGLSL3DRenderer *renderer,int textureType)
{
	if(NULL!=renderer && 0<=renderer->uniformTextureTypePos)
	{
		glUniform1i(renderer->uniformTextureTypePos,textureType);
		return YSOK;
	}
	return YSERR;
}

YSRESULT YsGLSLSet3DRendererTextureIdentifier(const struct YsGLSL3DRenderer *renderer,GLuint textureIdent)
{
	if(NULL!=renderer && 0<=renderer->uniformTexturePos)
	{
		glUniform1i(renderer->uniformTexturePos,textureIdent);
		return YSOK;
	}
	return YSERR;
}

YSRESULT YsGLSLSet3DRendererUniformTextureTilingMatrixfv(const struct YsGLSL3DRenderer *renderer,const GLfloat texTfm[16])
{
	if(NULL!=renderer && 0<=renderer->uniformTextureTileTransformPos)
	{
		glUniformMatrix4fv(renderer->uniformTextureTileTransformPos,1,GL_FALSE,texTfm);
		return YSOK;
	}
	return YSERR;
}

YSRESULT YsGLSLSet3DRendererBillBoardClippingType(const struct YsGLSL3DRenderer *renderer,int billBoardClippingType)
{
	if(NULL!=renderer && 0<=renderer->uniformBillBoardClippingTypePos)
	{
		glUniform1i(renderer->uniformBillBoardClippingTypePos,billBoardClippingType);
		return YSOK;
	}
	return YSERR;
}

YSRESULT YsGLSLSet3DRendererBillBoardTransformType(const struct YsGLSL3DRenderer *renderer,int billBoardTfmType)
{
	if(NULL!=renderer && 0<=renderer->uniformBillBoardTransformTypePos)
	{
		glUniform1i(renderer->uniformBillBoardTransformTypePos,billBoardTfmType);
		return YSOK;
	}
	return YSERR;
}

YSRESULT YsGLSLSet3DRendererBillBoardfv(const struct YsGLSL3DRenderer *renderer,const GLfloat center[3],const GLfloat dimension[2])
{
	if(NULL!=renderer && 0<=renderer->uniformBillBoardCenterPos && 0<=renderer->uniformBillBoardDimensionPos)
	{
		glUniform3fv(renderer->uniformBillBoardCenterPos,1,center);
		glUniform2fv(renderer->uniformBillBoardDimensionPos,1,dimension);
		return YSOK;
	}
	return YSERR;
}

YSRESULT YsGLSLSet3DRendererUniformLightDirectionfv(struct YsGLSL3DRenderer *renderer,int i,const GLfloat dir[3])
{
	if(NULL!=renderer && 0<=i && i<YSGLSLMAXNUMLIGHT && 0<=renderer->uniformLightPositionPos)
	{
		GLdouble l;
		GLdouble dir4d[4];
		dir4d[0]=dir[0];
		dir4d[1]=dir[1];
		dir4d[2]=dir[2];
		dir4d[3]=0.0;

		YsGLMultMatrixVectordv(dir4d,renderer->modelViewMatrix,dir4d);

		l=sqrt(dir4d[0]*dir4d[0]+dir4d[1]*dir4d[1]+dir4d[2]*dir4d[2]);
		renderer->lightPosition[i*4  ]=(GLfloat)(dir4d[0]/l);
		renderer->lightPosition[i*4+1]=(GLfloat)(dir4d[1]/l);
		renderer->lightPosition[i*4+2]=(GLfloat)(dir4d[2]/l);
		renderer->lightPosition[i*4+3]=0.0f;

		glUniform4fv(renderer->uniformLightPositionPos,YSGLSLMAXNUMLIGHT,renderer->lightPosition);
		return YSOK;
	}
	return YSERR;
}

YSRESULT YsGLSLSet3DRendererLightColor(struct YsGLSL3DRenderer *renderer,int idx,const GLfloat color[3])
{
	if(NULL!=renderer && 0<=idx && idx<YSGLSLMAXNUMLIGHT && 0<=renderer->uniformLightColorPos)
	{
		renderer->lightColor[idx*3  ]=color[0];
		renderer->lightColor[idx*3+1]=color[1];
		renderer->lightColor[idx*3+2]=color[2];
		glUniform3fv(renderer->uniformLightColorPos,YSGLSLMAXNUMLIGHT,renderer->lightColor);
		return YSOK;
	}
	return YSERR;
}

YSRESULT YsGLSLSet3DRendererAmbientColor(struct YsGLSL3DRenderer *renderer,const GLfloat color[3])
{
	if(NULL!=renderer && 0<=renderer->uniformAmbientColorPos)
	{
		renderer->ambientColor[0]=color[0];
		renderer->ambientColor[1]=color[1];
		renderer->ambientColor[2]=color[2];
		glUniform3fv(renderer->uniformAmbientColorPos,1,color);
		return YSOK;
	}
	return YSERR;
}

YSRESULT YsGLSLSet3DRendererSpecularColor(struct YsGLSL3DRenderer *renderer,const GLfloat color[3])
{
	if(NULL!=renderer && 0<=renderer->uniformSpecularColorPos)
	{
		renderer->specularColor[0]=color[0];
		renderer->specularColor[1]=color[1];
		renderer->specularColor[2]=color[2];
		glUniform3fv(renderer->uniformSpecularColorPos,1,color);
		return YSOK;
	}
	return YSERR;
}

YSRESULT YsGLSLSet3DRendererSpecularExponent(struct YsGLSL3DRenderer *renderer,const GLfloat exponent)
{
	if(NULL!=renderer && 0<=renderer->uniformSpecularExponentPos)
	{
		renderer->specularExponent=exponent;
		glUniform1f(renderer->uniformSpecularExponentPos,exponent);
		return YSOK;
	}
	return YSERR;
}

YSRESULT YsGLSLSet3DRendererUniformFlashSize(struct YsGLSL3DRenderer *renderer,GLfloat flashSize)
{
	if(NULL!=renderer && 0<=renderer->uniformDimensionPos)
	{
		glUniform1f(renderer->uniformDimensionPos,flashSize);
		return YSOK;
	}
	return YSERR;
}

YSRESULT YsGLSLSet3DRendererFlashRadius(struct YsGLSL3DRenderer *renderer,GLfloat r1,GLfloat r2)
{
	if(NULL!=renderer && 0<=renderer->uniformRadius1Pos && 0<=renderer->uniformRadius2Pos)
	{
		glUniform1f(renderer->uniformRadius1Pos,r1);
		glUniform1f(renderer->uniformRadius2Pos,r2);
		return YSOK;
	}
	return YSERR;
}

YSRESULT YsGLSLSet3DRendererUniformFogEnabled(struct YsGLSL3DRenderer *renderer,int enableFog)
{
	if(NULL!=renderer && 0<=renderer->uniformFogEnabledPos)
	{
		glUniform1i(renderer->uniformFogEnabledPos,enableFog);
		return YSOK;
	}
	return YSERR;
}

YSRESULT YsGLSLSet3DRendererUniformFogDensity(struct YsGLSL3DRenderer *renderer,const GLfloat density)
{
	if(NULL!=renderer && 0<=renderer->uniformFogDensityPos)
	{
		renderer->fogDensity=density;
		glUniform1f(renderer->uniformFogDensityPos,density);
		return YSOK;
	}
	return YSERR;
}

YSRESULT YsGLSLSet3DRendererUniformFogColor(struct YsGLSL3DRenderer *renderer,const GLfloat fogColor[4])
{
	if(NULL!=renderer && 0<=renderer->uniformFogColorPos)
	{
		glUniform4fv(renderer->uniformFogColorPos,1,fogColor);
		return YSOK;
	}
	return YSERR;
}

GLfloat YsGLSLGet3DRendererFogDensity(struct YsGLSL3DRenderer *renderer)
{
	if(NULL!=renderer)
	{
		return renderer->fogDensity;
	}
	return 0.0;
}

YSRESULT YsGLSLSet3DRendererZOffsetEnabled(struct YsGLSL3DRenderer *renderer,int enableZOffset)
{
	if(NULL!=renderer && 0<=renderer->uniformUseZOffsetPos)
	{
		glUniform1i(renderer->uniformUseZOffsetPos,enableZOffset);
		return YSOK;
	}
	return YSERR;
}

YSRESULT YsGLSLSet3DRendererZOffset(struct YsGLSL3DRenderer *renderer,GLfloat zOffset)
{
	if(NULL!=renderer && 0<=renderer->uniformZOffsetPos)
	{
		glUniform1f(renderer->uniformZOffsetPos,zOffset);
		return YSOK;
	}
	return YSERR;
}

GLuint YsGLSLUse3DRenderer(struct YsGLSL3DRenderer *renderer)
{
	GLuint prevProgramId;
	glGetIntegerv(GL_CURRENT_PROGRAM,(GLint *)&prevProgramId);

	if(NULL!=renderer)
	{
		if(1==renderer->inUse)
		{
			printf("Warning!!!! A 3DRenderer is activated again without being de-activated after last use!!!!\n");
		}

		glUseProgram(renderer->programId);

		if(0<=renderer->attribVertexPos)
		{
			glEnableVertexAttribArray(renderer->attribVertexPos);
		}
		if(0<=renderer->attribColorPos)
		{
			glEnableVertexAttribArray(renderer->attribColorPos);
		}
		if(0<=renderer->attribNormalPos)
		{
			glEnableVertexAttribArray(renderer->attribNormalPos);
		}
		if(0<=renderer->attribTexCoordPos)
		{
			glEnableVertexAttribArray(renderer->attribTexCoordPos);
		}
		if(0<=renderer->attribPixelOffsetPos)
		{
			glEnableVertexAttribArray(renderer->attribPixelOffsetPos);
		}
		if(0<=renderer->attribViewOffsetPos)
		{
			glEnableVertexAttribArray(renderer->attribViewOffsetPos);
		}

		renderer->inUse=1;
	}

	return prevProgramId;
}

void YsGLSLEndUse3DRenderer(struct YsGLSL3DRenderer *renderer)
{
	if(NULL!=renderer)
	{
		if(0<=renderer->attribVertexPos)
		{
			glDisableVertexAttribArray(renderer->attribVertexPos);
		}
		if(0<=renderer->attribColorPos)
		{
			glDisableVertexAttribArray(renderer->attribColorPos);
		}
		if(0<=renderer->attribNormalPos)
		{
			glDisableVertexAttribArray(renderer->attribNormalPos);
		}
		if(0<=renderer->attribTexCoordPos)
		{
			glDisableVertexAttribArray(renderer->attribTexCoordPos);
		}
		if(0<=renderer->attribPixelOffsetPos)
		{
			glDisableVertexAttribArray(renderer->attribPixelOffsetPos);
		}
		if(0<=renderer->attribViewOffsetPos)
		{
			glDisableVertexAttribArray(renderer->attribViewOffsetPos);
		}

		renderer->inUse=0;
	}
}

int YsGLSLCheck3DRendererIsInUse(struct YsGLSL3DRenderer *renderer)
{
	return renderer->inUse;
}

int YsGLSL3DRendererApplySphereMap(int w,int h,GLfloat *rgba,struct YsGLSL3DRenderer *renderer,int useSpecular)
{
	if(NULL!=renderer)
	{
		YsGLApplySphereMap(
		    w,h,rgba,
		    renderer->ambientColor,
		    YSGLSLMAXNUMLIGHT,
		    renderer->lightEnabled,
		    renderer->lightPosition,
		    renderer->lightColor,
		    useSpecular,
		    renderer->specularColor,
		    renderer->specularExponent);
		return YSOK;
	}
	return YSERR;
}

static void YsGLSL3DRendererInitializeUniform(struct YsGLSL3DRenderer *renderer)
{
	int i;
	GLuint prevProgramId;
	const GLfloat color[4]={1,1,1,1};
	const GLfloat fogColor[4]=
	{
		0,0,0,0
	};
	const GLfloat specularColor[3]={1,1,1};
	const GLfloat ambientColor[3]={0,0,0};

	const double identityMatrixd[16]=
	{
		1.0f,0.0f,0.0f,0.0f,
		0.0f,1.0f,0.0f,0.0f,
		0.0f,0.0f,1.0f,0.0f,
		0.0f,0.0f,0.0f,1.0f
	};

	glGetIntegerv(GL_CURRENT_PROGRAM,(GLint *)&prevProgramId);
	glUseProgram(renderer->programId);

	for(i=0; i<16; ++i)
	{
		renderer->modelViewMatrix[i]=identityMatrixd[i];
		renderer->projectionMatrix[i]=identityMatrixd[i];
	}

	YsGLSLLoadIdentityMatrix4(renderer->uniformProjectionPos);
	YsGLSLLoadIdentityMatrix4(renderer->uniformModelViewPos);
	if(0<=renderer->uniformColorPos)
	{
		glUniform4fv(renderer->uniformColorPos,1,color);
	}


	if(0<=renderer->uniformAlphaCutOffPos)
	{
		glUniform1f(renderer->uniformAlphaCutOffPos,0.0f);
	}


	if(0<=renderer->uniformTextureTypePos)
	{
		glUniform1i(renderer->uniformTextureTypePos,YSGLSL_TEX_TYPE_NONE);
	}
	if(0<=renderer->uniformUseNegativeOneToPositiveOnTexCoordPos)
	{
		glUniform1i(renderer->uniformUseNegativeOneToPositiveOnTexCoordPos,1);
	}
	if(0<=renderer->uniformTextureTileTransformPos)
	{
		YsGLSLLoadIdentityMatrix4(renderer->uniformTextureTileTransformPos);
	}
	if(0<=renderer->uniformTexturePos)
	{
		glUniform1i(renderer->uniformTexturePos,0);
	}

	if(0<=renderer->uniformBillBoardClippingTypePos)
	{
		glUniform1i(renderer->uniformBillBoardClippingTypePos,YSGLSL_BILLBOARD_CLIPPING_NONE);
	}
	if(0<=renderer->uniformBillBoardTransformTypePos)
	{
		glUniform1i(renderer->uniformBillBoardTransformTypePos,YSGLSL_TEX_BILLBOARD_PERS);
	}
	if(0<=renderer->uniformBillBoardCenterPos)
	{
		const GLfloat zero[3]={0.0f,0.0f,0.0f};
		glUniform3fv(renderer->uniformBillBoardCenterPos,1,zero);
	}
	if(0<=renderer->uniformBillBoardDimensionPos)
	{
		const GLfloat unit[2]={1.0f,1.0f};
		glUniform2fv(renderer->uniformBillBoardDimensionPos,1,unit);
	}

	if(0<=renderer->uniformFogEnabledPos)
	{
		glUniform1i(renderer->uniformFogEnabledPos,0);
	}
	if(0<=renderer->uniformFogDensityPos)
	{
		glUniform1f(renderer->uniformFogDensityPos,1.0f);
	}
	if(0<=renderer->uniformFogColorPos)
	{
		glUniform4fv(renderer->uniformFogColorPos,1,fogColor);
	}


	if(0<=renderer->uniformDimensionPos)
	{
		glUniform1f(renderer->uniformDimensionPos,0.5);
	}
	if(0<=renderer->uniformRadius1Pos)
	{
		glUniform1f(renderer->uniformRadius1Pos,0.2f);
	}
	if(0<=renderer->uniformRadius2Pos)
	{
		glUniform1f(renderer->uniformRadius2Pos,1.0f);
	}
	if(0<=renderer->uniformMaxIntensityPos)
	{
		glUniform1f(renderer->uniformMaxIntensityPos,1.1f);
	}
	if(0<=renderer->uniformAngleIntervalPos)
	{
		glUniform1f(renderer->uniformAngleIntervalPos,(GLfloat)YSGLPI/3.0);
	}
	if(0<=renderer->uniformAngleOffsetPos)
	{
		glUniform1f(renderer->uniformAngleOffsetPos,(GLfloat)YSGLPI/6.0);
	}
	if(0<=renderer->uniformExponentPos)
	{
		glUniform1f(renderer->uniformExponentPos,4);
	}


	for(i=0; i<YSGLSLMAXNUMLIGHT; i++)
	{
		renderer->lightEnabled[i]=(i==0 ? 1 : 0);
		renderer->lightPosition[i*4  ]=0;
		renderer->lightPosition[i*4+1]=0;
		renderer->lightPosition[i*4+2]=1;
		renderer->lightPosition[i*4+3]=0;
		renderer->lightColor[i*3  ]=1;
		renderer->lightColor[i*3+1]=1;
		renderer->lightColor[i*3+2]=1;
	}
	if(0<=renderer->uniformLightEnabledPos)
	{
		glUniform1iv(renderer->uniformLightEnabledPos,YSGLSLMAXNUMLIGHT,renderer->lightEnabled);
	}
	if(0<=renderer->uniformLightPositionPos)
	{
		glUniform4fv(renderer->uniformLightPositionPos,YSGLSLMAXNUMLIGHT,renderer->lightPosition);
	}
	if(0<=renderer->uniformLightColorPos)
	{
		glUniform3fv(renderer->uniformLightColorPos,YSGLSLMAXNUMLIGHT,renderer->lightColor);
	}
	if(0<=renderer->uniformColorPos)
	{
		glUniform4fv(renderer->uniformColorPos,1,color);
	}

	renderer->ambientColor[0]=ambientColor[0];
	renderer->ambientColor[1]=ambientColor[1];
	renderer->ambientColor[2]=ambientColor[2];
	if(0<=renderer->uniformAmbientColorPos)
	{
		glUniform3fv(renderer->uniformAmbientColorPos,1,ambientColor);
	}

	renderer->specularExponent=80.0f;
	if(0<=renderer->uniformSpecularExponentPos)
	{
		glUniform1f(renderer->uniformSpecularExponentPos,80.0f);
	}
	renderer->specularColor[0]=specularColor[0];
	renderer->specularColor[1]=specularColor[1];
	renderer->specularColor[2]=specularColor[2];
	if(0<=renderer->uniformSpecularColorPos)
	{
		glUniform3fv(renderer->uniformSpecularColorPos,1,specularColor);
	}


	if(0<=renderer->uniformUseZOffsetPos)
	{
		glUniform1i(renderer->uniformUseZOffsetPos,0);
	}
	if(0<=renderer->uniformZOffsetPos)
	{
		glUniform1f(renderer->uniformZOffsetPos,0.0f);
	}


	glUseProgram(prevProgramId);
}

#define YS_GLSL_VIEW_TRANSFORM_IN_VERTEX_SHADER \
	"	"YS_GLSL_HIGHP" vec4 vertexInView=modelView*vec4(vertex,1.0);\n"

#define YS_GLSL_PROJECTION_TRANSFORM_IN_VERTEX_SHADER \
	"	gl_Position=projection*vertexInView;\n"



/*
	YS_GLSL_STANDARD_LIGHTING
	Input variables: vec4 color
	                 vec3 nomLocal
	                 vec3 vecToCamera <- must be normalized
	Output variable: vec4 accumColor
*/
#define YS_GLSL_STANDARD_LIGHTING \
	"	"YS_GLSL_HIGHP" vec3 accumColor;\n", \
	"	accumColor=ambientColor*color.xyz;\n", \
	"\n", \
	"	for(int lightNo=0; lightNo<"YSGLSLMAXNUMLIGHTSTR"; lightNo++)\n", \
	"	{\n", \
	"		if(true==lightEnabled[lightNo])\n", \
	"		{\n", \
	"			"YS_GLSL_MIDP" float diffuseIntensity=max(dot(nomLocal,lightPos[lightNo].xyz),0.0);\n", \
	"			accumColor+=diffuseIntensity*(color.xyz*lightColor[lightNo]);\n", \
	"\n", \
	"			"YS_GLSL_HIGHP" vec3 mid=normalize(lightPos[lightNo].xyz+vecToCamera);\n", \
	"			"YS_GLSL_HIGHP" float specularIntensity=pow(max(dot(mid,nomLocal),0.0),specularExponent);\n", \
	"			accumColor+=specularIntensity*specularColor;\n", \
	"		}\n", \
	"	}\n"



#define YS_GLSL_ALPHA_CUTOFF_PARAMETER \
	"uniform "YS_GLSL_LOWP" float alphaCutOff;\n"

#define YS_GLSL_ALPHA_CUTOFF_IN_FRAGMENT_SHADER \
	"	if(gl_FragColor.a<alphaCutOff)\n", \
	"	{\n", \
	"		discard;\n", \
	"	}\n"


#define YS_GLSL_FOG_VARIABLES \
	"uniform bool fogEnabled;\n",\
	"uniform "YS_GLSL_MIDP" float fogDensity;\n",\
	"uniform "YS_GLSL_LOWP" vec4 fogColor;\n",\
	"varying "YS_GLSL_HIGHP" float zInViewCoord;\n"

#define YS_GLSL_FOG_CALCULATION_IN_VERTEXSHADER \
	"	if(false!=fogEnabled)\n",\
	"	{\n",\
	"		zInViewCoord=-vertexInView.z;\n",\
	"	}\n",\
	"	else\n",\
	"	{\n",\
	"		zInViewCoord=0.0;\n",\
	"	}\n"

#define YS_GLSL_FOGCALCULATION_IN_FRAGMENTSHADER \
	"	if(false!=fogEnabled)\n",\
	"	{\n",\
	"		"YS_GLSL_MIDP" float d=fogDensity*abs(zInViewCoord);\n",\
	"		"YS_GLSL_MIDP" float f=clamp(exp(-d*d),0.0,1.0);\n",\
	"		"YS_GLSL_LOWP" vec3 fogMix=mix(fogColor.rgb,gl_FragColor.rgb,f);\n",\
	"		gl_FragColor[0]=fogMix[0];\n",\
	"		gl_FragColor[1]=fogMix[1];\n",\
	"		gl_FragColor[2]=fogMix[2];\n",\
	"	}\n"

#define YS_GLSL_ZOFFSET_VARIABLE \
	"uniform bool useZOffset;\n",\
	"uniform "YS_GLSL_MIDP" float zOffset;\n"

#define YS_GLSL_ZOFFSET_IN_VERTEXSHADER \
	"	if(true==useZOffset)\n",\
	"	{\n",\
	"		gl_Position.z+=zOffset*gl_Position[3];\n",\
	"	}\n"


#define YS_GLSL_TEXTURE_VARIABLE \
	"uniform "YS_GLSL_LOWP" int textureType;\n", \
	"uniform "YS_GLSL_LOWP" mat4 textureTileTransform;\n", \
	"uniform int billBoardTransformType;\n", \
	"uniform bool useNegativeOneToPositiveOneTexCoord;\n", \
	"uniform "YS_GLSL_HIGHP" vec3 billBoardCenter;\n", \
	"uniform "YS_GLSL_HIGHP" vec2 billBoardDimension;\n", \
	"uniform sampler2D textureIdent;\n", \
	"varying "YS_GLSL_HIGHP" vec3 texCoordOut;\n"

// Memo: 
//   In orthographic projection, billBoardOffset can simply be calculated by taking XY components of 
//   billBoardOffsetInView-modelView*(vec4(billBoardCenter,1.0));
//   However, in perspective projection, a plane perpendicular to the viewing vector must be a reference plane,
//   and therefore heading and pitch angles must be calculated and taken into account.
//   Currently, this GLSL program only supports billboading in perspective projection.

#define YS_GLSL_TEXTURE_IN_VERTEXSHADER \
	"	if(textureType=="YSGLSL_TEX_TYPE_BILLBOARD_STR")\n", \
	"	{\n", \
	"		if("YSGLSL_TEX_BILLBOARD_ORTHO_STR"==billBoardTransformType)\n", \
	"		{\n", \
	"			"YS_GLSL_HIGHP" vec4 billBoardCenterInView=modelView*vec4(billBoardCenter,1.0);\n", \
	"			"YS_GLSL_MIDP" vec2 billBoardOffset=(vertexInView-billBoardCenterInView).xy;\n", \
	"			texCoordOut=vec3(billBoardOffset/billBoardDimension,0.0);\n", \
	"		}\n", \
	"		else\n", \
	"		{\n", \
	"			"YS_GLSL_HIGHP" vec4 billBoardCenterInView=modelView*vec4(billBoardCenter,1.0);\n", \
	"			"YS_GLSL_MIDP" vec3 viewVec=normalize(billBoardCenterInView.xyz);\n", \
	"			"YS_GLSL_MIDP" float pitch=asin(viewVec.y);\n", \
	"			"YS_GLSL_MIDP" float heading=-asin(viewVec.x);\n", \
	"			"YS_GLSL_MIDP" vec3 uVec=vec3(cos(heading),0.0,-sin(heading));\n", \
	"			"YS_GLSL_MIDP" vec3 vVec=vec3(-sin(pitch)*sin(heading),cos(pitch),sin(pitch)*cos(heading));\n", \
	"			"YS_GLSL_HIGHP" vec3 billBoardOffset3d=(vertexInView-billBoardCenterInView).xyz;\n", \
	"			"YS_GLSL_HIGHP" vec2 billBoardOffset=vec2(dot(billBoardOffset3d,uVec),dot(billBoardOffset3d,vVec));\n", \
	"			texCoordOut=vec3(billBoardOffset/billBoardDimension,0.0);\n", \
	"		}\n", \
	"		if(true==useNegativeOneToPositiveOneTexCoord)\n", \
	"		{\n", \
	"			texCoordOut=(texCoordOut+vec3(1.0,1.0,1.0))/vec3(2.0,2.0,2.0);\n", \
	"		}\n", \
	"	}\n", \
	"	else if(textureType=="YSGLSL_TEX_TYPE_TILING_STR" || textureType=="YSGLSL_TEX_TYPE_3DTILING_STR")\n", \
	"	{\n", \
	"		"YS_GLSL_LOWP" vec4 texCoord=textureTileTransform*vec4(vertex,1.0);\n", \
	"		texCoordOut=texCoord.xyz;\n", \
	"		if(true==useNegativeOneToPositiveOneTexCoord)\n", \
	"		{\n", \
	"			texCoordOut=(texCoordOut+vec3(1.0,1.0,1.0))/vec3(2.0,2.0,2.0);\n", \
	"		}\n", \
	"	}\n", \
	"	else\n", \
	"	{\n", \
	"		texCoordOut[0]=0.0;\n", \
	"		texCoordOut[1]=0.0;\n", \
	"		texCoordOut[2]=0.0;\n", \
	"	}\n"

#define YS_GLSL_TEXTURE_IN_FRAGMENTSHADER \
	"	if("YSGLSL_TEX_TYPE_NONE_STR"==textureType)\n", \
	"	{\n", \
	"	}\n", \
	"	else if("YSGLSL_TEX_TYPE_TILING_STR"==textureType ||\n", \
	"	        "YSGLSL_TEX_TYPE_BILLBOARD_STR"==textureType)\n", \
	"	{\n", \
	"		"YS_GLSL_LOWP" vec4 texcell=texture2D(textureIdent,texCoordOut.xy);\n", \
	"		gl_FragColor=gl_FragColor*texcell;\n", \
	"		//gl_FragColor.r*=texcell.a;\n",\
	"		//gl_FragColor.g*=texcell.a;\n",\
	"		//gl_FragColor.b*=texcell.a;\n",\
	"	}\n", \
	"	else if("YSGLSL_TEX_TYPE_3DTILING_STR"==textureType)\n", \
	"	{\n", \
	"		"YS_GLSL_LOWP" vec4 texcell[3],avg;\n", \
	"		texcell[0]=texture2D(textureIdent,texCoordOut.xy);\n", \
	"		texcell[1]=texture2D(textureIdent,texCoordOut.xz);\n", \
	"		texcell[2]=texture2D(textureIdent,texCoordOut.yz);\n", \
	"		avg=texcell[0]/3.0+texcell[1]/3.0+texcell[2]/3.0;\n", \
	"		gl_FragColor=gl_FragColor*avg;\n", \
	"	}\n", \
	"\n"


////////////////////////////////////////////////////////////
// Flat
static const char *flat3DDrawingVertexShader[]=
{
	"uniform mat4 projection;\n",
	"uniform mat4 modelView;\n",
	"uniform "YS_GLSL_LOWP" vec4 color;\n",
	"\n",
	"attribute "YS_GLSL_HIGHP" vec3 vertex;\n",
	"\n",
	YS_GLSL_TEXTURE_VARIABLE,
	YS_GLSL_FOG_VARIABLES,
	YS_GLSL_ZOFFSET_VARIABLE,
	"\n",
	"void main()\n",
	"{\n",
	YS_GLSL_VIEW_TRANSFORM_IN_VERTEX_SHADER,
	YS_GLSL_TEXTURE_IN_VERTEXSHADER,
	YS_GLSL_FOG_CALCULATION_IN_VERTEXSHADER,
	YS_GLSL_PROJECTION_TRANSFORM_IN_VERTEX_SHADER,
	YS_GLSL_ZOFFSET_IN_VERTEXSHADER,
	"}\n",
	"\n",
};

static const char *flat3DDrawingFragmentShader[]=
{
	"uniform "YS_GLSL_LOWP" vec4 color;\n",
	"\n",
	YS_GLSL_TEXTURE_VARIABLE,
	YS_GLSL_FOG_VARIABLES,
	YS_GLSL_ALPHA_CUTOFF_PARAMETER,
	"\n",
	"void main()\n",
	"{\n",
	"	gl_FragColor=color;\n",
	YS_GLSL_TEXTURE_IN_FRAGMENTSHADER,
	YS_GLSL_ALPHA_CUTOFF_IN_FRAGMENT_SHADER,
	YS_GLSL_FOGCALCULATION_IN_FRAGMENTSHADER,
	"\n",
	"}\n",
	"\n",
};

static void YsGLSLFlat3DRendererCreateProgram(struct YsGLSL3DRenderer *renderer)
{
	renderer->vertexShaderId=glCreateShader(GL_VERTEX_SHADER);
	renderer->fragmentShaderId=glCreateShader(GL_FRAGMENT_SHADER);
	renderer->programId=glCreateProgram();

	YsGLSLCompileAndLinkVertexAndFragmentShader(
	    renderer->programId,
	    renderer->vertexShaderId,
	    sizeof(flat3DDrawingVertexShader)/sizeof(const char *),
	    flat3DDrawingVertexShader,
	    renderer->fragmentShaderId,
	    sizeof(flat3DDrawingFragmentShader)/sizeof(const char *),
	    flat3DDrawingFragmentShader);

	// Uniforms
	renderer->uniformProjectionPos=glGetUniformLocation(renderer->programId,"projection");
	renderer->uniformModelViewPos=glGetUniformLocation(renderer->programId,"modelView");
	renderer->uniformColorPos=glGetUniformLocation(renderer->programId,"color");


	renderer->uniformFogEnabledPos=glGetUniformLocation(renderer->programId,"fogEnabled");
	renderer->uniformFogDensityPos=glGetUniformLocation(renderer->programId,"fogDensity");
	renderer->uniformFogColorPos=glGetUniformLocation(renderer->programId,"fogColor");

	renderer->uniformUseZOffsetPos=glGetUniformLocation(renderer->programId,"useZOffset");
	renderer->uniformZOffsetPos=glGetUniformLocation(renderer->programId,"zOffset");

	renderer->uniformAlphaCutOffPos=glGetUniformLocation(renderer->programId,"alphaCutOff");

	YsGLSLGetTextureUniformPos(renderer);


	// Attributes
	renderer->attribVertexPos=glGetAttribLocation(renderer->programId,"vertex");
}

struct YsGLSL3DRenderer *YsGLSLCreateFlat3DRenderer(void)
{
	struct YsGLSL3DRenderer *renderer=YsGLSLCreateAndInitialize3DRenderer();
	if(NULL!=renderer)
	{
		YsGLSLFlat3DRendererCreateProgram(renderer);
		YsGLSL3DRendererInitializeUniform(renderer);
	}
	return renderer;
}

YSRESULT YsGLSLDrawPrimitiveVtxfv(const struct YsGLSL3DRenderer *renderer,GLenum mode,GLsizei nVtx,const GLfloat vtx[])
{
	if(NULL!=renderer &&
	   0<=renderer->attribVertexPos &&
	   0>renderer->attribColorPos &&
	   0>renderer->attribNormalPos &&
	   0>renderer->attribTexCoordPos)
	{
		glVertexAttribPointer(renderer->attribVertexPos,3,GL_FLOAT,GL_FALSE,0,vtx);
		glDrawArrays(mode,0,nVtx);
		return YSOK;
	}
	else
	{
		return YSERR;
	}
}


////////////////////////////////////////////////////////////
// MonoColorBillBoard
static const char *monoColorBillBoard3DDrawingVertexShader[]=
{
	"uniform mat4 projection;\n",
	"uniform mat4 modelView;\n",
	"uniform "YS_GLSL_LOWP" vec4 color;\n",
	"uniform int billBoardTransformType;\n",
	"uniform bool useNegativeOneToPositiveOneTexCoord;\n",
	"uniform sampler2D textureIdent;\n",
	"\n",
	"attribute "YS_GLSL_HIGHP" vec3 vertex;\n",
	"attribute "YS_GLSL_MIDP" vec3 viewOffset;\n",
	"attribute "YS_GLSL_MIDP" vec2 texCoord;\n",
	"\n",
	"varying "YS_GLSL_HIGHP" vec2 texCoordOut;\n",
	"\n",
	YS_GLSL_FOG_VARIABLES,
	YS_GLSL_ZOFFSET_VARIABLE,
	"\n",
	"void main()\n",
	"{\n",
		YS_GLSL_VIEW_TRANSFORM_IN_VERTEX_SHADER,
	"\n",
	"	if("YSGLSL_TEX_BILLBOARD_ORTHO_STR"==billBoardTransformType)\n",
	"	{\n",
	"		vertexInView.xyz+=viewOffset;\n",
	"	}\n",
	"	else\n",
	"	{\n",
	"		vec3 axis[3];\n",
	"		"YS_GLSL_MIDP" vec3 viewVec=normalize(vertexInView.xyz);\n",
	"		"YS_GLSL_MIDP" float pitch=asin(viewVec.y);\n",
	"		"YS_GLSL_MIDP" float heading=-asin(viewVec.x);\n",
	"		axis[0]=vec3(cos(heading),0.0,-sin(heading));\n",
	"		axis[1]=vec3(-sin(pitch)*sin(heading),cos(pitch),sin(pitch)*cos(heading));\n",
	"		axis[2]=-viewVec;\n",
	"		vertexInView.xyz+=(axis[0]*viewOffset[0]+axis[1]*viewOffset[1]+axis[2]*viewOffset[2]);\n",
	"	}\n",
	"	texCoordOut=texCoord;\n",
	"	if(true==useNegativeOneToPositiveOneTexCoord)\n",
	"	{\n",
	"		texCoordOut=(texCoordOut+vec2(1.0,1.0))/vec2(2.0,2.0);\n",
	"	}\n",
	"\n",
		YS_GLSL_FOG_CALCULATION_IN_VERTEXSHADER,
		YS_GLSL_PROJECTION_TRANSFORM_IN_VERTEX_SHADER,
		YS_GLSL_ZOFFSET_IN_VERTEXSHADER,
	"}\n",
	"\n",
};

static const char *monoColorBillBoard3DDrawingFragmentShader[]=
{
	"uniform "YS_GLSL_LOWP" vec4 color;\n",
	"\n",
	"uniform int billBoardClippingType;\n",
	"uniform int billBoardTransformType;\n", \
	"uniform sampler2D textureIdent;\n", \
	"\n",
	"varying "YS_GLSL_HIGHP" vec2 texCoordOut;\n"
	"\n",
	YS_GLSL_FOG_VARIABLES,
	YS_GLSL_ALPHA_CUTOFF_PARAMETER,
	"\n",
	"void main()\n",
	"{\n",
	"	if("YSGLSL_BILLBOARD_CLIPPING_CIRCLE_STR"==billBoardClippingType)\n",
	"	{\n",
	"		"YS_GLSL_MIDP" vec2 nomLocal=(texCoordOut.xy*2.0)-vec2(1.0,1.0);\n",
	"		if(1.0<=dot(nomLocal,nomLocal))\n",
	"		{\n",
	"			discard;\n",
	"		}\n",
	"	}\n",
	"\n",
	"	"YS_GLSL_LOWP" vec4 texcell=texture2D(textureIdent,texCoordOut);\n", \
	"	gl_FragColor=color*texcell;\n",
	"\n",
		YS_GLSL_ALPHA_CUTOFF_IN_FRAGMENT_SHADER,
		YS_GLSL_FOGCALCULATION_IN_FRAGMENTSHADER,
	"}\n",
	"\n",
};

static void YsGLSLMonoColorBillBoard3DRendererCreateProgram(struct YsGLSL3DRenderer *renderer)
{
	renderer->vertexShaderId=glCreateShader(GL_VERTEX_SHADER);
	renderer->fragmentShaderId=glCreateShader(GL_FRAGMENT_SHADER);
	renderer->programId=glCreateProgram();

	YsGLSLCompileAndLinkVertexAndFragmentShader(
	    renderer->programId,
	    renderer->vertexShaderId,
	    sizeof(monoColorBillBoard3DDrawingVertexShader)/sizeof(const char *),
	    monoColorBillBoard3DDrawingVertexShader,
	    renderer->fragmentShaderId,
	    sizeof(monoColorBillBoard3DDrawingFragmentShader)/sizeof(const char *),
	    monoColorBillBoard3DDrawingFragmentShader);

	// Uniforms
	renderer->uniformProjectionPos=glGetUniformLocation(renderer->programId,"projection");
	renderer->uniformModelViewPos=glGetUniformLocation(renderer->programId,"modelView");
	renderer->uniformColorPos=glGetUniformLocation(renderer->programId,"color");
	renderer->uniformBillBoardClippingTypePos=glGetUniformLocation(renderer->programId,"billBoardClippingType");
	renderer->uniformBillBoardTransformTypePos=glGetUniformLocation(renderer->programId,"billBoardTransformType");
	renderer->uniformTexturePos=glGetUniformLocation(renderer->programId,"textureIdent");
	renderer->uniformUseNegativeOneToPositiveOnTexCoordPos=glGetUniformLocation(renderer->programId,"useNegativeOneToPositiveOneTexCoord");

	renderer->uniformLightEnabledPos=glGetUniformLocation(renderer->programId,"lightEnabled");
	renderer->uniformLightPositionPos=glGetUniformLocation(renderer->programId,"lightPos");
	renderer->uniformLightColorPos=glGetUniformLocation(renderer->programId,"lightColor");
	renderer->uniformAmbientColorPos=glGetUniformLocation(renderer->programId,"ambientColor");


	renderer->uniformFogEnabledPos=glGetUniformLocation(renderer->programId,"fogEnabled");
	renderer->uniformFogDensityPos=glGetUniformLocation(renderer->programId,"fogDensity");
	renderer->uniformFogColorPos=glGetUniformLocation(renderer->programId,"fogColor");

	renderer->uniformUseZOffsetPos=glGetUniformLocation(renderer->programId,"useZOffset");
	renderer->uniformZOffsetPos=glGetUniformLocation(renderer->programId,"zOffset");

	renderer->uniformAlphaCutOffPos=glGetUniformLocation(renderer->programId,"alphaCutOff");


	// Attributes
	renderer->attribVertexPos=glGetAttribLocation(renderer->programId,"vertex");
	renderer->attribTexCoordPos=glGetAttribLocation(renderer->programId,"texCoord");
	renderer->attribViewOffsetPos=glGetAttribLocation(renderer->programId,"viewOffset");
}

struct YsGLSL3DRenderer *YsGLSLCreateMonoColorBillBoard3DRenderer(void)
{
	struct YsGLSL3DRenderer *renderer=YsGLSLCreateAndInitialize3DRenderer();
	if(NULL!=renderer)
	{
		YsGLSLMonoColorBillBoard3DRendererCreateProgram(renderer);
		YsGLSL3DRendererInitializeUniform(renderer);
	}
	return renderer;
}


////////////////////////////////////////////////////////////
// VariColorBillBoard
static const char *variColorBillBoard3DDrawingVertexShader[]=
{
	"uniform mat4 projection;\n",
	"uniform mat4 modelView;\n",
	"uniform int billBoardTransformType;\n",
	"uniform bool useNegativeOneToPositiveOneTexCoord;\n",
	"uniform sampler2D textureIdent;\n",
	"\n",
	"attribute "YS_GLSL_HIGHP" vec3 vertex;\n",
	"attribute "YS_GLSL_LOWP" vec4 colorIn;\n",
	"attribute "YS_GLSL_MIDP" vec3 viewOffset;\n",
	"attribute "YS_GLSL_MIDP" vec2 texCoord;\n",
	"\n",
	"varying "YS_GLSL_HIGHP" vec2 texCoordOut;\n",
	"varying "YS_GLSL_LOWP" vec4 color;\n",
	YS_GLSL_FOG_VARIABLES,
	YS_GLSL_ZOFFSET_VARIABLE,
	"\n",
	"void main()\n",
	"{\n",
	YS_GLSL_VIEW_TRANSFORM_IN_VERTEX_SHADER,
	"\n",
	"	if("YSGLSL_TEX_BILLBOARD_ORTHO_STR"==billBoardTransformType)\n",
	"	{\n",
	"		vertexInView.xyz+=viewOffset;\n",
	"	}\n",
	"	else\n",
	"	{\n",
	"		vec3 axis[3];\n",
	"		"YS_GLSL_MIDP" vec3 viewVec=normalize(vertexInView.xyz);\n",
	"		"YS_GLSL_MIDP" float pitch=asin(viewVec.y);\n",
	"		"YS_GLSL_MIDP" float heading=-asin(viewVec.x);\n",
	"		axis[0]=vec3(cos(heading),0.0,-sin(heading));\n",
	"		axis[1]=vec3(-sin(pitch)*sin(heading),cos(pitch),sin(pitch)*cos(heading));\n",
	"		axis[2]=-viewVec;\n",
	"		vertexInView.xyz+=(axis[0]*viewOffset[0]+axis[1]*viewOffset[1]+axis[2]*viewOffset[2]);\n",
	"	}\n",
	"	texCoordOut=texCoord;\n",
	"	if(true==useNegativeOneToPositiveOneTexCoord)\n",
	"	{\n",
	"		texCoordOut=(texCoordOut+vec2(1.0,1.0))/vec2(2.0,2.0);\n",
	"	}\n",
	"	color=colorIn;\n",
	"\n",
		YS_GLSL_FOG_CALCULATION_IN_VERTEXSHADER,
		YS_GLSL_PROJECTION_TRANSFORM_IN_VERTEX_SHADER,
		YS_GLSL_ZOFFSET_IN_VERTEXSHADER,
	"}\n",
	"\n",
};

static const char *variColorBillBoard3DDrawingFragmentShader[]=
{
	"uniform int billBoardClippingType;\n",
	"uniform int billBoardTransformType;\n", \
	"uniform sampler2D textureIdent;\n", \
	"\n",
	"varying "YS_GLSL_HIGHP" vec2 texCoordOut;\n"
	"varying "YS_GLSL_LOWP" vec4 color;\n",
	"\n",
	YS_GLSL_FOG_VARIABLES,
	YS_GLSL_ALPHA_CUTOFF_PARAMETER,
	"\n",
	"void main()\n",
	"{\n",
	"	if("YSGLSL_BILLBOARD_CLIPPING_CIRCLE_STR"==billBoardClippingType)\n",
	"	{\n",
	"		"YS_GLSL_MIDP" vec2 nomLocal=(texCoordOut.xy*2.0)-vec2(1.0,1.0);\n",
	"		if(1.0<=dot(nomLocal,nomLocal))\n",
	"		{\n",
	"			discard;\n",
	"		}\n",
	"	}\n",
	"\n",
	"	"YS_GLSL_LOWP" vec4 texcell=texture2D(textureIdent,texCoordOut);\n", \
	"	gl_FragColor=color*texcell;\n",
	"\n",
	YS_GLSL_ALPHA_CUTOFF_IN_FRAGMENT_SHADER,
	YS_GLSL_FOGCALCULATION_IN_FRAGMENTSHADER,
	"}\n",
	"\n",
};

static void YsGLSLVariColorBillBoard3DRendererCreateProgram(struct YsGLSL3DRenderer *renderer)
{
	renderer->vertexShaderId=glCreateShader(GL_VERTEX_SHADER);
	renderer->fragmentShaderId=glCreateShader(GL_FRAGMENT_SHADER);
	renderer->programId=glCreateProgram();

	YsGLSLCompileAndLinkVertexAndFragmentShader(
	    renderer->programId,
	    renderer->vertexShaderId,
	    sizeof(variColorBillBoard3DDrawingVertexShader)/sizeof(const char *),
	    variColorBillBoard3DDrawingVertexShader,
	    renderer->fragmentShaderId,
	    sizeof(variColorBillBoard3DDrawingFragmentShader)/sizeof(const char *),
	    variColorBillBoard3DDrawingFragmentShader);

	// Uniforms
	renderer->uniformProjectionPos=glGetUniformLocation(renderer->programId,"projection");
	renderer->uniformModelViewPos=glGetUniformLocation(renderer->programId,"modelView");
	renderer->uniformBillBoardClippingTypePos=glGetUniformLocation(renderer->programId,"billBoardClippingType");
	renderer->uniformBillBoardTransformTypePos=glGetUniformLocation(renderer->programId,"billBoardTransformType");
	renderer->uniformTexturePos=glGetUniformLocation(renderer->programId,"textureIdent");
	renderer->uniformUseNegativeOneToPositiveOnTexCoordPos=glGetUniformLocation(renderer->programId,"useNegativeOneToPositiveOneTexCoord");

	renderer->uniformLightEnabledPos=glGetUniformLocation(renderer->programId,"lightEnabled");
	renderer->uniformLightPositionPos=glGetUniformLocation(renderer->programId,"lightPos");
	renderer->uniformLightColorPos=glGetUniformLocation(renderer->programId,"lightColor");
	renderer->uniformSpecularColorPos=glGetUniformLocation(renderer->programId,"specularColor");
	renderer->uniformSpecularExponentPos=glGetUniformLocation(renderer->programId,"specularExponent");
	renderer->uniformAmbientColorPos=glGetUniformLocation(renderer->programId,"ambientColor");


	renderer->uniformFogEnabledPos=glGetUniformLocation(renderer->programId,"fogEnabled");
	renderer->uniformFogDensityPos=glGetUniformLocation(renderer->programId,"fogDensity");
	renderer->uniformFogColorPos=glGetUniformLocation(renderer->programId,"fogColor");

	renderer->uniformUseZOffsetPos=glGetUniformLocation(renderer->programId,"useZOffset");
	renderer->uniformZOffsetPos=glGetUniformLocation(renderer->programId,"zOffset");

	renderer->uniformAlphaCutOffPos=glGetUniformLocation(renderer->programId,"alphaCutOff");


	// Attributes
	renderer->attribVertexPos=glGetAttribLocation(renderer->programId,"vertex");
	renderer->attribColorPos=glGetAttribLocation(renderer->programId,"colorIn");
	renderer->attribTexCoordPos=glGetAttribLocation(renderer->programId,"texCoord");
	renderer->attribViewOffsetPos=glGetAttribLocation(renderer->programId,"viewOffset");
}

struct YsGLSL3DRenderer *YsGLSLCreateVariColorBillBoard3DRenderer(void)
{
	struct YsGLSL3DRenderer *renderer=YsGLSLCreateAndInitialize3DRenderer();
	if(NULL!=renderer)
	{
		YsGLSLVariColorBillBoard3DRendererCreateProgram(renderer);
		YsGLSL3DRendererInitializeUniform(renderer);
	}
	return renderer;
}

YSRESULT YsGLSLDrawPrimitiveVtxColVoffsetTexcoordfv(
    const struct YsGLSL3DRenderer *renderer,GLenum mode,
    GLsizei nVtx,
    const GLfloat vtx[],     // 3*nVtx
    const GLfloat col[],     // 4*nVtx
    const GLfloat vOffset[], // 3*nVtx   Let it modify Z as well
    const GLfloat texCoord[] // 2*nVtx
    )
{
	if(NULL!=renderer &&
	   0<=renderer->attribVertexPos &&
	   0<=renderer->attribColorPos &&
	   0<=renderer->attribViewOffsetPos &&
	   0<=renderer->attribTexCoordPos)
	{
		glVertexAttribPointer(renderer->attribVertexPos,3,GL_FLOAT,GL_FALSE,0,vtx);
		glVertexAttribPointer(renderer->attribColorPos,4,GL_FLOAT,GL_FALSE,0,col);
		glVertexAttribPointer(renderer->attribViewOffsetPos,3,GL_FLOAT,GL_FALSE,0,vOffset);
		glVertexAttribPointer(renderer->attribTexCoordPos,2,GL_FLOAT,GL_FALSE,0,texCoord);
		glDrawArrays(mode,0,nVtx);
		return YSOK;
	}
	else
	{
		return YSERR;
	}
}


YSRESULT YsGLSLDrawPrimitiveVtxVoffsetTexcoordfv(
    const struct YsGLSL3DRenderer *renderer,GLenum mode,
    GLsizei nVtx,
    const GLfloat vtx[],     // 3*nVtx
    const GLfloat vOffset[], // 3*nVtx   Let it modify Z as well
    const GLfloat texCoord[] // 2*nVtx
    )
{
	if(NULL!=renderer &&
	   0<=renderer->attribVertexPos &&
	   0<=renderer->attribViewOffsetPos &&
	   0<=renderer->attribTexCoordPos)
	{
		glVertexAttribPointer(renderer->attribVertexPos,3,GL_FLOAT,GL_FALSE,0,vtx);
		glVertexAttribPointer(renderer->attribViewOffsetPos,3,GL_FLOAT,GL_FALSE,0,vOffset);
		glVertexAttribPointer(renderer->attribTexCoordPos,2,GL_FLOAT,GL_FALSE,0,texCoord);
		glDrawArrays(mode,0,nVtx);
		return YSOK;
	}
	else
	{
		return YSERR;
	}
}


////////////////////////////////////////////////////////////
// Flash
static const char *flash3DDrawingVertexShader[]=
{
	"uniform mat4 projection;\n",
	"uniform mat4 modelView;\n",
	"uniform "YS_GLSL_MIDP" float dimension;\n",
	"uniform "YS_GLSL_MIDP" float radius1;\n",
	"uniform "YS_GLSL_MIDP" float radius2;\n",
	"uniform "YS_GLSL_MIDP" float maxIntensity;\n",
	"uniform "YS_GLSL_MIDP" float angleInterval;\n",
	"uniform "YS_GLSL_MIDP" float angleOffset;\n",
	"uniform "YS_GLSL_MIDP" float exponent;\n",
	"\n",
	"attribute "YS_GLSL_HIGHP" vec3 vertex;\n",
	"attribute "YS_GLSL_LOWP" vec4 colorIn;\n",
	"attribute "YS_GLSL_HIGHP" vec3 viewOffset;\n",
	"\n",
	"varying "YS_GLSL_LOWP" vec2 uv;\n",
	"varying "YS_GLSL_LOWP" vec4 color;\n",
	"\n",
	YS_GLSL_ZOFFSET_VARIABLE,
	YS_GLSL_FOG_VARIABLES,
	"\n",
	"void main()\n",
	"{\n",
	YS_GLSL_VIEW_TRANSFORM_IN_VERTEX_SHADER,
	"   vec3 offset=viewOffset*dimension;\n",
	"   vertexInView.x+=offset.x;\n",
	"   vertexInView.y+=offset.y;\n",
	"   vertexInView.z+=offset.z;\n",
	"\n",
	YS_GLSL_FOG_CALCULATION_IN_VERTEXSHADER,
	YS_GLSL_PROJECTION_TRANSFORM_IN_VERTEX_SHADER,
	"   uv=vec2(viewOffset.x,viewOffset.y);\n",
	"	color=colorIn;\n",
	YS_GLSL_ZOFFSET_IN_VERTEXSHADER,
	"}\n",
	"\n",
};

static const char *flash3DDrawingFragmentShader[]=
{
	"uniform "YS_GLSL_MIDP" float radius1;\n",
	"uniform "YS_GLSL_MIDP" float radius2;\n",
	"uniform "YS_GLSL_MIDP" float maxIntensity;\n",
	"uniform "YS_GLSL_MIDP" float angleInterval;\n",
	"uniform "YS_GLSL_MIDP" float angleOffset;\n",
	"uniform "YS_GLSL_MIDP" float exponent;\n",
	"\n",
	"varying "YS_GLSL_LOWP" vec2 uv;\n",
	"varying "YS_GLSL_LOWP" vec4 color;\n",
	"\n",
	// No alpha cut off in the flash renderer.
	YS_GLSL_FOG_VARIABLES,
	"\n",
	"void main()\n",
	"{\n",
	"    const "YS_GLSL_MIDP" float PI=3.1415927;\n",
	"    const "YS_GLSL_MIDP" float halfPI=3.1415927/2.0;\n",
	"    const "YS_GLSL_MIDP" float twoPI=3.1415927*2.0;\n",
	"    "YS_GLSL_MIDP" float r=length(uv);\n",
	"    if(r<=radius1)\n",
	"    {\n",
	"        "YS_GLSL_MIDP" float param=halfPI*(radius1-r)/radius1;\n",
	"        "YS_GLSL_MIDP" float additive=sin(param)*maxIntensity;\n",
	"        gl_FragColor[0]=min(1.0,color[0]+additive);\n",
	"        gl_FragColor[1]=min(1.0,color[1]+additive);\n",
	"        gl_FragColor[2]=min(1.0,color[2]+additive);\n",
	"        gl_FragColor[3]=color[3];\n",
	YS_GLSL_FOGCALCULATION_IN_FRAGMENTSHADER,
	"    }\n",
	"    else\n",
	"    {\n",
	"		 "YS_GLSL_MIDP" float halfInterval=angleInterval/2.0;\n",
	"        "YS_GLSL_MIDP" float theata=twoPI+angleOffset+atan(uv.y,uv.x);\n",
	"        "YS_GLSL_MIDP" float t=mod(theata,angleInterval);\n",
	"        "YS_GLSL_MIDP" float s=1.0-abs(t-halfInterval)/halfInterval;\n",
	"\n",
	"        "YS_GLSL_MIDP" float rf=radius1+(radius2-radius1)*pow(s,exponent);\n",
	"        if(rf<r)\n",
	"	     {\n",
	"            discard;\n",
	"        }\n",
	"        else\n",
	"        {\n",
	"            "YS_GLSL_MIDP" float alpha=max(1.0-(r-radius1)/(rf-radius1),0.0);\n",
	"            gl_FragColor[0]=color[0];\n",
	"            gl_FragColor[1]=color[1];\n",
	"            gl_FragColor[2]=color[2];\n",
	"            gl_FragColor[3]=color[3]*alpha;\n",
	YS_GLSL_FOGCALCULATION_IN_FRAGMENTSHADER,
	"        }\n",
	"    }\n",
	"}\n",
	"\n",
};

static void YsGLSLFlash3DRendererCreateProgram(struct YsGLSL3DRenderer *renderer)
{
	renderer->vertexShaderId=glCreateShader(GL_VERTEX_SHADER);
	renderer->fragmentShaderId=glCreateShader(GL_FRAGMENT_SHADER);
	renderer->programId=glCreateProgram();

	YsGLSLCompileAndLinkVertexAndFragmentShader(
	    renderer->programId,
	    renderer->vertexShaderId,
	    sizeof(flash3DDrawingVertexShader)/sizeof(const char *),
	    flash3DDrawingVertexShader,
	    renderer->fragmentShaderId,
	    sizeof(flash3DDrawingFragmentShader)/sizeof(const char *),
	    flash3DDrawingFragmentShader);

	// Uniforms
	renderer->uniformProjectionPos=glGetUniformLocation(renderer->programId,"projection");
	renderer->uniformModelViewPos=glGetUniformLocation(renderer->programId,"modelView");

	renderer->uniformDimensionPos=glGetUniformLocation(renderer->programId,"dimension");
	renderer->uniformRadius1Pos=glGetUniformLocation(renderer->programId,"radius1");
	renderer->uniformRadius2Pos=glGetUniformLocation(renderer->programId,"radius2");
	renderer->uniformMaxIntensityPos=glGetUniformLocation(renderer->programId,"maxIntensity");
	renderer->uniformAngleIntervalPos=glGetUniformLocation(renderer->programId,"angleInterval");
	renderer->uniformAngleOffsetPos=glGetUniformLocation(renderer->programId,"angleOffset");
	renderer->uniformExponentPos=glGetUniformLocation(renderer->programId,"exponent");

	renderer->uniformFogEnabledPos=glGetUniformLocation(renderer->programId,"fogEnabled");
	renderer->uniformFogDensityPos=glGetUniformLocation(renderer->programId,"fogDensity");
	renderer->uniformFogColorPos=glGetUniformLocation(renderer->programId,"fogColor");

	renderer->uniformUseZOffsetPos=glGetUniformLocation(renderer->programId,"useZOffset");
	renderer->uniformZOffsetPos=glGetUniformLocation(renderer->programId,"zOffset");


	// Attributes
	renderer->attribVertexPos=glGetAttribLocation(renderer->programId,"vertex");
	renderer->attribViewOffsetPos=glGetAttribLocation(renderer->programId,"viewOffset");
	renderer->attribColorPos=glGetAttribLocation(renderer->programId,"colorIn");
}

struct YsGLSL3DRenderer *YsGLSLCreateFlash3DRenderer(void)
{
	struct YsGLSL3DRenderer *renderer=YsGLSLCreateAndInitialize3DRenderer();
	if(NULL!=renderer)
	{
		YsGLSLFlash3DRendererCreateProgram(renderer);
		YsGLSL3DRendererInitializeUniform(renderer);
	}
	return renderer;
}


YSRESULT YsGLSLDrawPrimitiveVtxColVoffsetfv(const struct YsGLSL3DRenderer *renderer,GLsizei nVtx,const GLfloat vtx[],const GLfloat col[],const GLfloat viewOffset[])
{
	if(NULL!=renderer &&
	   0<=renderer->attribViewOffsetPos &&
	   0<=renderer->attribVertexPos && 
	   0<=renderer->attribColorPos)
	{
		glVertexAttribPointer(renderer->attribVertexPos,3,GL_FLOAT,GL_FALSE,0,vtx);
		glVertexAttribPointer(renderer->attribViewOffsetPos,3,GL_FLOAT,GL_FALSE,0,viewOffset);
		glVertexAttribPointer(renderer->attribColorPos,4,GL_FLOAT,GL_FALSE,0,col);
		glDrawArrays(GL_TRIANGLES,0,nVtx);
		return YSOK;
	}
	else
	{
		return YSERR;
	}
}


////////////////////////////////////////////////////////////
// Vari color

static const char *variColor3DDrawingVertexShader[]=
{
	"uniform mat4 projection;\n",
	"uniform mat4 modelView;\n",
	"\n",
	"attribute "YS_GLSL_HIGHP" vec3 vertex;\n",
	"attribute "YS_GLSL_LOWP" vec4 colorIn;\n",
	"\n",
	"varying "YS_GLSL_LOWP" vec4 color;\n",
	"\n",
	YS_GLSL_TEXTURE_VARIABLE,
	YS_GLSL_ZOFFSET_VARIABLE,
	YS_GLSL_FOG_VARIABLES,
	"\n",
	"void main()\n",
	"{\n",
	YS_GLSL_VIEW_TRANSFORM_IN_VERTEX_SHADER,
	YS_GLSL_PROJECTION_TRANSFORM_IN_VERTEX_SHADER,
	"\n",
	"	color=colorIn;\n",
	"\n",
	YS_GLSL_TEXTURE_IN_VERTEXSHADER,
	YS_GLSL_FOG_CALCULATION_IN_VERTEXSHADER,
	YS_GLSL_ZOFFSET_IN_VERTEXSHADER,
	"\n",
	"}\n",
	"\n"
};

static const char *variColor3DDrawingFragmentShader[]=
{
	"varying "YS_GLSL_LOWP" vec4 color;\n",
	"\n",
	YS_GLSL_TEXTURE_VARIABLE,
	YS_GLSL_FOG_VARIABLES,
	YS_GLSL_ALPHA_CUTOFF_PARAMETER,
	"\n",
	"void main()\n",
	"{\n",
	"	gl_FragColor=color;\n",
	YS_GLSL_TEXTURE_IN_FRAGMENTSHADER,
	YS_GLSL_ALPHA_CUTOFF_IN_FRAGMENT_SHADER,
	YS_GLSL_FOGCALCULATION_IN_FRAGMENTSHADER,
	"}\n",
	"\n"
};

static void YsGLSLVariColor3DRendererCreateProgram(struct YsGLSL3DRenderer *renderer)
{
	renderer->vertexShaderId=glCreateShader(GL_VERTEX_SHADER);
	renderer->fragmentShaderId=glCreateShader(GL_FRAGMENT_SHADER);
	renderer->programId=glCreateProgram();

	YsGLSLCompileAndLinkVertexAndFragmentShader(
	    renderer->programId,
	    renderer->vertexShaderId,
	    sizeof(variColor3DDrawingVertexShader)/sizeof(const char *),
	    variColor3DDrawingVertexShader,
	    renderer->fragmentShaderId,
	    sizeof(variColor3DDrawingFragmentShader)/sizeof(const char *),
	    variColor3DDrawingFragmentShader);

	// Uniforms
	renderer->uniformProjectionPos=glGetUniformLocation(renderer->programId,"projection");
	renderer->uniformModelViewPos=glGetUniformLocation(renderer->programId,"modelView");

	renderer->uniformFogEnabledPos=glGetUniformLocation(renderer->programId,"fogEnabled");
	renderer->uniformFogDensityPos=glGetUniformLocation(renderer->programId,"fogDensity");
	renderer->uniformFogColorPos=glGetUniformLocation(renderer->programId,"fogColor");

	renderer->uniformUseZOffsetPos=glGetUniformLocation(renderer->programId,"useZOffset");
	renderer->uniformZOffsetPos=glGetUniformLocation(renderer->programId,"zOffset");

	renderer->uniformAlphaCutOffPos=glGetUniformLocation(renderer->programId,"alphaCutOff");

	YsGLSLGetTextureUniformPos(renderer);


	// Attributes
	renderer->attribVertexPos=glGetAttribLocation(renderer->programId,"vertex");
	renderer->attribColorPos=glGetAttribLocation(renderer->programId,"colorIn");
}

struct YsGLSL3DRenderer *YsGLSLCreateVariColor3DRenderer(void)
{
	struct YsGLSL3DRenderer *renderer=YsGLSLCreateAndInitialize3DRenderer();
	if(NULL!=renderer)
	{
		YsGLSLVariColor3DRendererCreateProgram(renderer);
		YsGLSL3DRendererInitializeUniform(renderer);
	}
	return renderer;
}

YSRESULT YsGLSLDrawPrimitiveVtxColfv(const struct YsGLSL3DRenderer *renderer,GLenum mode,GLsizei nVtx,const GLfloat vtx[],const GLfloat col[])
{
	if(NULL!=renderer &&
	   0<=renderer->attribVertexPos &&
	   0<=renderer->attribColorPos &&
	   0>renderer->attribNormalPos &&
	   0>renderer->attribTexCoordPos &&
	   0>renderer->attribPixelOffsetPos)
	{
		glVertexAttribPointer(renderer->attribVertexPos,3,GL_FLOAT,GL_FALSE,0,vtx);
		glVertexAttribPointer(renderer->attribColorPos,4,GL_FLOAT,GL_FALSE,0,col);
		glDrawArrays(mode,0,nVtx);
		return YSOK;
	}
	else
	{
		return YSERR;
	}
}

////////////////////////////////////////////////////////////
// MonoColorPerVtxShading

static const char *monoColorPerVtxShading3DDrawingVertexShader[]=
{
	"uniform mat4 projection;\n",
	"uniform mat4 modelView;\n",
	"\n",
	"uniform bool lightEnabled["YSGLSLMAXNUMLIGHTSTR"];\n",
	"uniform vec4 lightPos["YSGLSLMAXNUMLIGHTSTR"];\n",
	"uniform vec3 lightColor["YSGLSLMAXNUMLIGHTSTR"];\n",
	"uniform vec3 specularColor;\n",
	"uniform float specularExponent;\n",
	"uniform vec3 ambientColor;\n",
	"uniform "YS_GLSL_LOWP"vec4 color;\n",
	"\n",
	"attribute "YS_GLSL_HIGHP" vec3 vertex;\n",
	"attribute "YS_GLSL_HIGHP" vec3 normal;\n",
	"\n",
	"varying "YS_GLSL_LOWP" vec4 colorOut;\n",
	"\n",
	YS_GLSL_TEXTURE_VARIABLE,
	YS_GLSL_ZOFFSET_VARIABLE,
	YS_GLSL_FOG_VARIABLES,
	"\n",
	"void main()\n",
	"{\n",
	YS_GLSL_VIEW_TRANSFORM_IN_VERTEX_SHADER,
	YS_GLSL_PROJECTION_TRANSFORM_IN_VERTEX_SHADER,
	"\n",
	"	"YS_GLSL_HIGHP" vec3 nomLocal=normalize((modelView*vec4(normal,0)).xyz);\n",
	"	"YS_GLSL_HIGHP" vec3 vecToCamera=-normalize(vertexInView.xyz);\n",
	"\n",
	YS_GLSL_STANDARD_LIGHTING,
	"\n",
	"	colorOut=vec4(accumColor,color[3]);\n",
	"\n",
	YS_GLSL_TEXTURE_IN_VERTEXSHADER,
	YS_GLSL_FOG_CALCULATION_IN_VERTEXSHADER,
	YS_GLSL_ZOFFSET_IN_VERTEXSHADER,
	"\n",
	"}\n",
	"\n"
};

static const char *monoColorPerVtxShading3DDrawingFragmentShader[]=
{
	"varying "YS_GLSL_LOWP" vec4 colorOut;\n",
	"\n",
	YS_GLSL_TEXTURE_VARIABLE,
	YS_GLSL_FOG_VARIABLES,
	YS_GLSL_ALPHA_CUTOFF_PARAMETER,
	"\n",
	"void main()\n",
	"{\n",
	"	gl_FragColor=colorOut;\n",
	YS_GLSL_TEXTURE_IN_FRAGMENTSHADER,
	YS_GLSL_ALPHA_CUTOFF_IN_FRAGMENT_SHADER,
	YS_GLSL_FOGCALCULATION_IN_FRAGMENTSHADER,
	"}\n",
	"\n"
};

static void YsGLSLMonoColorPerVtxShading3DRendererCreateProgram(struct YsGLSL3DRenderer *renderer)
{
	renderer->vertexShaderId=glCreateShader(GL_VERTEX_SHADER);
	renderer->fragmentShaderId=glCreateShader(GL_FRAGMENT_SHADER);
	renderer->programId=glCreateProgram();

	YsGLSLCompileAndLinkVertexAndFragmentShader(
	    renderer->programId,
	    renderer->vertexShaderId,
	    sizeof(monoColorPerVtxShading3DDrawingVertexShader)/sizeof(const char *),
	    monoColorPerVtxShading3DDrawingVertexShader,
	    renderer->fragmentShaderId,
	    sizeof(monoColorPerVtxShading3DDrawingFragmentShader)/sizeof(const char *),
	    monoColorPerVtxShading3DDrawingFragmentShader);

	// Uniforms
	renderer->uniformProjectionPos=glGetUniformLocation(renderer->programId,"projection");
	renderer->uniformModelViewPos=glGetUniformLocation(renderer->programId,"modelView");

	renderer->uniformLightEnabledPos=glGetUniformLocation(renderer->programId,"lightEnabled");
	renderer->uniformLightPositionPos=glGetUniformLocation(renderer->programId,"lightPos");
	renderer->uniformLightColorPos=glGetUniformLocation(renderer->programId,"lightColor");
	renderer->uniformSpecularColorPos=glGetUniformLocation(renderer->programId,"specularColor");
	renderer->uniformSpecularExponentPos=glGetUniformLocation(renderer->programId,"specularExponent");
	renderer->uniformAmbientColorPos=glGetUniformLocation(renderer->programId,"ambientColor");
	renderer->uniformColorPos=glGetUniformLocation(renderer->programId,"color");

	renderer->uniformFogEnabledPos=glGetUniformLocation(renderer->programId,"fogEnabled");
	renderer->uniformFogDensityPos=glGetUniformLocation(renderer->programId,"fogDensity");
	renderer->uniformFogColorPos=glGetUniformLocation(renderer->programId,"fogColor");

	renderer->uniformUseZOffsetPos=glGetUniformLocation(renderer->programId,"useZOffset");
	renderer->uniformZOffsetPos=glGetUniformLocation(renderer->programId,"zOffset");

	renderer->uniformAlphaCutOffPos=glGetUniformLocation(renderer->programId,"alphaCutOff");

	YsGLSLGetTextureUniformPos(renderer);

	// Attributes
	renderer->attribVertexPos=glGetAttribLocation(renderer->programId,"vertex");
	renderer->attribNormalPos=glGetAttribLocation(renderer->programId,"normal");
}

struct YsGLSL3DRenderer *YsGLSLCreateMonoColorPerVtxShading3DRenderer(void)
{
	struct YsGLSL3DRenderer *renderer=YsGLSLCreateAndInitialize3DRenderer();
	if(NULL!=renderer)
	{
		YsGLSLMonoColorPerVtxShading3DRendererCreateProgram(renderer);
		YsGLSL3DRendererInitializeUniform(renderer);
	}
	return renderer;
}

YSRESULT YsGLSLDrawPrimitiveVtxNomfv(const struct YsGLSL3DRenderer *renderer,GLenum mode,GLsizei nVtx,const GLfloat vtx[],const GLfloat nom[])
{
	if(NULL!=renderer &&
	   0<=renderer->attribVertexPos &&
	   0>renderer->attribColorPos &&
	   0<=renderer->attribNormalPos &&
	   0>renderer->attribTexCoordPos &&
	   0>renderer->attribPixelOffsetPos)
	{
		glVertexAttribPointer(renderer->attribVertexPos,3,GL_FLOAT,GL_FALSE,0,vtx);
		glVertexAttribPointer(renderer->attribNormalPos,3,GL_FLOAT,GL_FALSE,0,nom);
		glDrawArrays(mode,0,nVtx);
		return YSOK;
	}
	else
	{
		return YSERR;
	}
}


////////////////////////////////////////////////////////////
// MonoColorPerPixShading

static const char *monoColorPerPixShading3DDrawingVertexShader[]=
{
	"uniform bool lightEnabled["YSGLSLMAXNUMLIGHTSTR"];\n",
	"uniform "YS_GLSL_HIGHP"vec4 lightPos["YSGLSLMAXNUMLIGHTSTR"];\n",
	"uniform "YS_GLSL_LOWP"vec3 lightColor["YSGLSLMAXNUMLIGHTSTR"];\n",
	"uniform "YS_GLSL_LOWP"vec3 specularColor;\n",
	"uniform "YS_GLSL_LOWP"float specularExponent;\n",
	"uniform "YS_GLSL_LOWP"vec3 ambientColor;\n",
	"uniform "YS_GLSL_LOWP"vec4 color;\n",
	"\n",
	"uniform "YS_GLSL_HIGHP"mat4 projection;\n",
	"uniform "YS_GLSL_HIGHP"mat4 modelView;\n",
	"\n",
	"attribute "YS_GLSL_HIGHP" vec3 vertex;\n",
	"attribute "YS_GLSL_HIGHP" vec3 normal;\n",
	"\n",
	"varying "YS_GLSL_HIGHP" vec3 normalOut;\n",
	"varying "YS_GLSL_HIGHP" vec3 vecToCameraOut;\n",
	"\n",
	YS_GLSL_TEXTURE_VARIABLE,
	YS_GLSL_ZOFFSET_VARIABLE,
	YS_GLSL_FOG_VARIABLES,
	"\n",
	"void main()\n",
	"{\n",
	YS_GLSL_VIEW_TRANSFORM_IN_VERTEX_SHADER,
	YS_GLSL_PROJECTION_TRANSFORM_IN_VERTEX_SHADER,
	"\n",
	"	"YS_GLSL_HIGHP"vec4 normal4d=vec4(normal,0);\n",
	"	normalOut=normalize((modelView*normal4d).xyz);\n",
	"	vecToCameraOut=-normalize(vertexInView.xyz);\n",
	"\n",
	YS_GLSL_TEXTURE_IN_VERTEXSHADER,
	YS_GLSL_FOG_CALCULATION_IN_VERTEXSHADER,
	YS_GLSL_ZOFFSET_IN_VERTEXSHADER,
	"\n",
	"}\n",
	"\n"
};

static const char *monoColorPerPixShading3DDrawingFragmentShader[]=
{
	"uniform bool lightEnabled["YSGLSLMAXNUMLIGHTSTR"];\n",
	"uniform "YS_GLSL_HIGHP"vec4 lightPos["YSGLSLMAXNUMLIGHTSTR"];\n",
	"uniform "YS_GLSL_LOWP"vec3 lightColor["YSGLSLMAXNUMLIGHTSTR"];\n",
	"uniform "YS_GLSL_LOWP"vec3 specularColor;\n",
	"uniform "YS_GLSL_LOWP"float specularExponent;\n",
	"uniform "YS_GLSL_LOWP"vec3 ambientColor;\n",
	"uniform "YS_GLSL_LOWP"vec4 color;\n",
	"\n",
	"varying "YS_GLSL_HIGHP" vec3 normalOut;\n",
	"varying "YS_GLSL_HIGHP" vec3 vecToCameraOut;\n",
	"\n",
	YS_GLSL_TEXTURE_VARIABLE,
	YS_GLSL_ALPHA_CUTOFF_PARAMETER,
	YS_GLSL_FOG_VARIABLES,
	"\n",
	"void main()\n",
	"{\n",
	"	"YS_GLSL_LOWP"vec3 nomLocal=normalize(normalOut);\n",
	"	"YS_GLSL_LOWP"vec3 vecToCamera=normalize(vecToCameraOut);\n",
	"\n",
	YS_GLSL_STANDARD_LIGHTING,
	"\n",
	"	gl_FragColor=vec4(accumColor,color[3]);\n",
	YS_GLSL_TEXTURE_IN_FRAGMENTSHADER,
	YS_GLSL_ALPHA_CUTOFF_IN_FRAGMENT_SHADER,
	YS_GLSL_FOGCALCULATION_IN_FRAGMENTSHADER,
	"}\n",
	"\n"
};

static void YsGLSLMonoColorPerPixShading3DRendererCreateProgram(struct YsGLSL3DRenderer *renderer)
{
	renderer->vertexShaderId=glCreateShader(GL_VERTEX_SHADER);
	renderer->fragmentShaderId=glCreateShader(GL_FRAGMENT_SHADER);
	renderer->programId=glCreateProgram();

	YsGLSLCompileAndLinkVertexAndFragmentShader(
	    renderer->programId,
	    renderer->vertexShaderId,
	    sizeof(monoColorPerPixShading3DDrawingVertexShader)/sizeof(const char *),
	    monoColorPerPixShading3DDrawingVertexShader,
	    renderer->fragmentShaderId,
	    sizeof(monoColorPerPixShading3DDrawingFragmentShader)/sizeof(const char *),
	    monoColorPerPixShading3DDrawingFragmentShader);

	// Uniforms
	renderer->uniformProjectionPos=glGetUniformLocation(renderer->programId,"projection");
	renderer->uniformModelViewPos=glGetUniformLocation(renderer->programId,"modelView");

	renderer->uniformLightEnabledPos=glGetUniformLocation(renderer->programId,"lightEnabled");
	renderer->uniformLightPositionPos=glGetUniformLocation(renderer->programId,"lightPos");
	renderer->uniformLightColorPos=glGetUniformLocation(renderer->programId,"lightColor");
	renderer->uniformSpecularColorPos=glGetUniformLocation(renderer->programId,"specularColor");
	renderer->uniformSpecularExponentPos=glGetUniformLocation(renderer->programId,"specularExponent");
	renderer->uniformAmbientColorPos=glGetUniformLocation(renderer->programId,"ambientColor");
	renderer->uniformColorPos=glGetUniformLocation(renderer->programId,"color");

	renderer->uniformFogEnabledPos=glGetUniformLocation(renderer->programId,"fogEnabled");
	renderer->uniformFogDensityPos=glGetUniformLocation(renderer->programId,"fogDensity");
	renderer->uniformFogColorPos=glGetUniformLocation(renderer->programId,"fogColor");

	renderer->uniformUseZOffsetPos=glGetUniformLocation(renderer->programId,"useZOffset");
	renderer->uniformZOffsetPos=glGetUniformLocation(renderer->programId,"zOffset");

	renderer->uniformAlphaCutOffPos=glGetUniformLocation(renderer->programId,"alphaCutOff");

	YsGLSLGetTextureUniformPos(renderer);

	// Attributes
	renderer->attribVertexPos=glGetAttribLocation(renderer->programId,"vertex");
	renderer->attribNormalPos=glGetAttribLocation(renderer->programId,"normal");
}

struct YsGLSL3DRenderer *YsGLSLCreateMonoColorPerPixShading3DRenderer(void)
{
	struct YsGLSL3DRenderer *renderer=YsGLSLCreateAndInitialize3DRenderer();
	if(NULL!=renderer)
	{
		YsGLSLMonoColorPerPixShading3DRendererCreateProgram(renderer);
		YsGLSL3DRendererInitializeUniform(renderer);
	}
	return renderer;
}

////////////////////////////////////////////////////////////
// VariColorPerVtxShading
static const char *variColorPerVtxShading3DDrawingVertexShader[]=
{
	"uniform "YS_GLSL_HIGHP"mat4 projection;\n",
	"uniform "YS_GLSL_HIGHP"mat4 modelView;\n",
	"\n",
	"uniform bool lightEnabled["YSGLSLMAXNUMLIGHTSTR"];\n",
	"uniform "YS_GLSL_HIGHP"vec4 lightPos["YSGLSLMAXNUMLIGHTSTR"];\n",
	"uniform "YS_GLSL_LOWP"vec3 lightColor["YSGLSLMAXNUMLIGHTSTR"];\n",
	"uniform "YS_GLSL_LOWP"vec3 specularColor;\n",
	"uniform "YS_GLSL_LOWP"float specularExponent;\n",
	"uniform "YS_GLSL_LOWP"vec3 ambientColor;\n",
	"\n",
	"attribute "YS_GLSL_HIGHP" vec3 vertex;\n",
	"attribute "YS_GLSL_HIGHP" vec3 normal;\n",
	"attribute "YS_GLSL_LOWP" vec4 color;\n",    // <- Only difference from monoColorPerVtxShader
	"\n",
	"varying "YS_GLSL_LOWP" vec4 colorOut;\n",
	"\n",
	YS_GLSL_TEXTURE_VARIABLE,
	YS_GLSL_ZOFFSET_VARIABLE,
	YS_GLSL_FOG_VARIABLES,
	"\n",
	"void main()\n",
	"{\n",
	YS_GLSL_VIEW_TRANSFORM_IN_VERTEX_SHADER,
	YS_GLSL_PROJECTION_TRANSFORM_IN_VERTEX_SHADER,
	"\n",
	"	"YS_GLSL_HIGHP" vec3 nomLocal=normalize((modelView*vec4(normal,0)).xyz);\n",
	"	"YS_GLSL_HIGHP" vec3 vecToCamera=-normalize(vertexInView.xyz);\n",
	"\n",
	YS_GLSL_STANDARD_LIGHTING,
	"\n",
	"	colorOut=vec4(accumColor,color[3]);\n",
	"\n",
	YS_GLSL_TEXTURE_IN_VERTEXSHADER,
	YS_GLSL_FOG_CALCULATION_IN_VERTEXSHADER,
	YS_GLSL_ZOFFSET_IN_VERTEXSHADER,
	"\n",
	"}\n",
	"\n"
};

static const char *variColorPerVtxShading3DDrawingFragmentShader[]=
{
	"varying "YS_GLSL_LOWP" vec4 colorOut;\n",
	"\n",
	YS_GLSL_TEXTURE_VARIABLE,
	YS_GLSL_ALPHA_CUTOFF_PARAMETER,
	YS_GLSL_FOG_VARIABLES,
	"\n",
	"void main()\n",
	"{\n",
	"	gl_FragColor=colorOut;\n",
	YS_GLSL_TEXTURE_IN_FRAGMENTSHADER,
	YS_GLSL_ALPHA_CUTOFF_IN_FRAGMENT_SHADER,
	YS_GLSL_FOGCALCULATION_IN_FRAGMENTSHADER,
	"}\n",
	"\n"
};

static void YsGLSLVariColorPerVtxShading3DRendererCreateProgram(struct YsGLSL3DRenderer *renderer)
{
	renderer->vertexShaderId=glCreateShader(GL_VERTEX_SHADER);
	renderer->fragmentShaderId=glCreateShader(GL_FRAGMENT_SHADER);
	renderer->programId=glCreateProgram();

	YsGLSLCompileAndLinkVertexAndFragmentShader(
	    renderer->programId,
	    renderer->vertexShaderId,
	    sizeof(variColorPerVtxShading3DDrawingVertexShader)/sizeof(const char *),
	    variColorPerVtxShading3DDrawingVertexShader,
	    renderer->fragmentShaderId,
	    sizeof(variColorPerVtxShading3DDrawingFragmentShader)/sizeof(const char *),
	    variColorPerVtxShading3DDrawingFragmentShader);

	// Uniforms
	renderer->uniformProjectionPos=glGetUniformLocation(renderer->programId,"projection");
	renderer->uniformModelViewPos=glGetUniformLocation(renderer->programId,"modelView");

	renderer->uniformLightEnabledPos=glGetUniformLocation(renderer->programId,"lightEnabled");
	renderer->uniformLightPositionPos=glGetUniformLocation(renderer->programId,"lightPos");
	renderer->uniformLightColorPos=glGetUniformLocation(renderer->programId,"lightColor");
	renderer->uniformSpecularColorPos=glGetUniformLocation(renderer->programId,"specularColor");
	renderer->uniformSpecularExponentPos=glGetUniformLocation(renderer->programId,"specularExponent");
	renderer->uniformAmbientColorPos=glGetUniformLocation(renderer->programId,"ambientColor");

	renderer->uniformFogEnabledPos=glGetUniformLocation(renderer->programId,"fogEnabled");
	renderer->uniformFogDensityPos=glGetUniformLocation(renderer->programId,"fogDensity");
	renderer->uniformFogColorPos=glGetUniformLocation(renderer->programId,"fogColor");

	renderer->uniformUseZOffsetPos=glGetUniformLocation(renderer->programId,"useZOffset");
	renderer->uniformZOffsetPos=glGetUniformLocation(renderer->programId,"zOffset");

	renderer->uniformAlphaCutOffPos=glGetUniformLocation(renderer->programId,"alphaCutOff");

	YsGLSLGetTextureUniformPos(renderer);

	// Attributes
	renderer->attribVertexPos=glGetAttribLocation(renderer->programId,"vertex");
	renderer->attribNormalPos=glGetAttribLocation(renderer->programId,"normal");
	renderer->attribColorPos=glGetAttribLocation(renderer->programId,"color");
}

struct YsGLSL3DRenderer *YsGLSLCreateVariColorPerVtxShading3DRenderer(void)
{
	struct YsGLSL3DRenderer *renderer=YsGLSLCreateAndInitialize3DRenderer();
	if(NULL!=renderer)
	{
		YsGLSLVariColorPerVtxShading3DRendererCreateProgram(renderer);
		YsGLSL3DRendererInitializeUniform(renderer);
	}
	return renderer;
}

YSRESULT YsGLSLDrawPrimitiveVtxNomColfv(const struct YsGLSL3DRenderer *renderer, GLenum mode,GLsizei nVtx,const GLfloat vtx[],const GLfloat nom[],const GLfloat col[])
{
	if(NULL!=renderer &&
	   0<=renderer->attribVertexPos &&
	   0<=renderer->attribColorPos &&
	   0<=renderer->attribNormalPos &&
	   0>renderer->attribTexCoordPos &&
	   0>renderer->attribPixelOffsetPos)
	{
		glVertexAttribPointer(renderer->attribVertexPos,3,GL_FLOAT,GL_FALSE,0,vtx);
		glVertexAttribPointer(renderer->attribNormalPos,3,GL_FLOAT,GL_FALSE,0,nom);
		glVertexAttribPointer(renderer->attribColorPos,4,GL_FLOAT,GL_FALSE,0,col);
		glDrawArrays(mode,0,nVtx);
		return YSOK;
	}
	else if(0<=renderer->attribVertexPos &&
	        0>renderer->attribColorPos &&
	        0<=renderer->attribNormalPos &&
	        0>renderer->attribTexCoordPos &&
	        0>renderer->attribPixelOffsetPos &&
	        0<=renderer->uniformColorPos)
	{
		YsGLSLSet3DRendererUniformColorfv(renderer,col);
		glVertexAttribPointer(renderer->attribVertexPos,3,GL_FLOAT,GL_FALSE,0,vtx);
		glVertexAttribPointer(renderer->attribNormalPos,3,GL_FLOAT,GL_FALSE,0,nom);
		glDrawArrays(mode,0,nVtx);
		return YSOK;
	}
	else
	{
		return YSERR;
	}
}

////////////////////////////////////////////////////////////
// VariColorPerPixShading

static const char *variColorPerPixShading3DDrawingVertexShader[]=
{
	"uniform "YS_GLSL_HIGHP"mat4 projection;\n",
	"uniform "YS_GLSL_HIGHP"mat4 modelView;\n",
	"\n",
	"attribute "YS_GLSL_HIGHP" vec3 vertex;\n",
	"attribute "YS_GLSL_HIGHP" vec3 normal;\n",
	"attribute "YS_GLSL_LOWP"vec4 colorIn;\n",
	"\n",
	"varying "YS_GLSL_HIGHP" vec3 normalOut;\n",
	"varying "YS_GLSL_HIGHP" vec3 vecToCameraOut;\n",
	"varying "YS_GLSL_LOWP" vec4 color;\n",
	"\n",
	YS_GLSL_TEXTURE_VARIABLE,
	YS_GLSL_ZOFFSET_VARIABLE,
	YS_GLSL_FOG_VARIABLES,
	"\n",
	"void main()\n",
	"{\n",
	YS_GLSL_VIEW_TRANSFORM_IN_VERTEX_SHADER,
	YS_GLSL_PROJECTION_TRANSFORM_IN_VERTEX_SHADER,
	"\n",
	"	"YS_GLSL_HIGHP"vec4 normal4d=vec4(normal,0);\n",
	"	normalOut=normalize((modelView*normal4d).xyz);\n",
	"	vecToCameraOut=-normalize(vertexInView.xyz);\n",
	"\n",
	"	color=colorIn;\n",
	"\n",
	YS_GLSL_TEXTURE_IN_VERTEXSHADER,
	YS_GLSL_FOG_CALCULATION_IN_VERTEXSHADER,
	YS_GLSL_ZOFFSET_IN_VERTEXSHADER,
	"\n",
	"}\n",
	"\n"
};

static const char *variColorPerPixShading3DDrawingFragmentShader[]=
{
	"uniform bool lightEnabled["YSGLSLMAXNUMLIGHTSTR"];\n",
	"uniform "YS_GLSL_HIGHP"vec4 lightPos["YSGLSLMAXNUMLIGHTSTR"];\n",
	"uniform "YS_GLSL_LOWP"vec3 lightColor["YSGLSLMAXNUMLIGHTSTR"];\n",
	"uniform "YS_GLSL_LOWP"vec3 specularColor;\n",
	"uniform "YS_GLSL_LOWP"float specularExponent;\n",
	"uniform "YS_GLSL_LOWP"vec3 ambientColor;\n",
	"\n",
	"varying "YS_GLSL_HIGHP" vec3 normalOut;\n",
	"varying "YS_GLSL_HIGHP" vec3 vecToCameraOut;\n",
	"varying "YS_GLSL_LOWP" vec4 color;\n",
	"\n",
	YS_GLSL_TEXTURE_VARIABLE,
	YS_GLSL_ALPHA_CUTOFF_PARAMETER,
	YS_GLSL_FOG_VARIABLES,
	"\n",
	"void main()\n",
	"{\n",
	"	"YS_GLSL_HIGHP"vec3 nomLocal=normalize(normalOut);\n",
	"	"YS_GLSL_HIGHP"vec3 vecToCamera=normalize(vecToCameraOut);\n",
	"\n",
	YS_GLSL_STANDARD_LIGHTING,
	"\n",
	"	gl_FragColor=vec4(accumColor,color[3]);\n",
	"\n",
	YS_GLSL_TEXTURE_IN_FRAGMENTSHADER,
	YS_GLSL_ALPHA_CUTOFF_IN_FRAGMENT_SHADER,
	YS_GLSL_FOGCALCULATION_IN_FRAGMENTSHADER,
	"}\n",
	"\n"
};

static void YsGLSLVariColorPerPixShading3DRendererCreateProgram(struct YsGLSL3DRenderer *renderer)
{
	renderer->vertexShaderId=glCreateShader(GL_VERTEX_SHADER);
	renderer->fragmentShaderId=glCreateShader(GL_FRAGMENT_SHADER);
	renderer->programId=glCreateProgram();

	YsGLSLCompileAndLinkVertexAndFragmentShader(
	    renderer->programId,
	    renderer->vertexShaderId,
	    sizeof(variColorPerPixShading3DDrawingVertexShader)/sizeof(const char *),
	    variColorPerPixShading3DDrawingVertexShader,
	    renderer->fragmentShaderId,
	    sizeof(variColorPerPixShading3DDrawingFragmentShader)/sizeof(const char *),
	    variColorPerPixShading3DDrawingFragmentShader);

	// Uniforms
	renderer->uniformProjectionPos=glGetUniformLocation(renderer->programId,"projection");
	renderer->uniformModelViewPos=glGetUniformLocation(renderer->programId,"modelView");

	renderer->uniformLightEnabledPos=glGetUniformLocation(renderer->programId,"lightEnabled");
	renderer->uniformLightPositionPos=glGetUniformLocation(renderer->programId,"lightPos");
	renderer->uniformLightColorPos=glGetUniformLocation(renderer->programId,"lightColor");
	renderer->uniformSpecularColorPos=glGetUniformLocation(renderer->programId,"specularColor");
	renderer->uniformSpecularExponentPos=glGetUniformLocation(renderer->programId,"specularExponent");
	renderer->uniformAmbientColorPos=glGetUniformLocation(renderer->programId,"ambientColor");

	renderer->uniformFogEnabledPos=glGetUniformLocation(renderer->programId,"fogEnabled");
	renderer->uniformFogDensityPos=glGetUniformLocation(renderer->programId,"fogDensity");
	renderer->uniformFogColorPos=glGetUniformLocation(renderer->programId,"fogColor");

	renderer->uniformUseZOffsetPos=glGetUniformLocation(renderer->programId,"useZOffset");
	renderer->uniformZOffsetPos=glGetUniformLocation(renderer->programId,"zOffset");

	renderer->uniformAlphaCutOffPos=glGetUniformLocation(renderer->programId,"alphaCutOff");

	YsGLSLGetTextureUniformPos(renderer);


	// Attributes
	renderer->attribVertexPos=glGetAttribLocation(renderer->programId,"vertex");
	renderer->attribNormalPos=glGetAttribLocation(renderer->programId,"normal");
	renderer->attribColorPos=glGetAttribLocation(renderer->programId,"colorIn");
}

struct YsGLSL3DRenderer *YsGLSLCreateVariColorPerPixShading3DRenderer(void)
{
	struct YsGLSL3DRenderer *renderer=YsGLSLCreateAndInitialize3DRenderer();
	if(NULL!=renderer)
	{
		YsGLSLVariColorPerPixShading3DRendererCreateProgram(renderer);
		YsGLSL3DRendererInitializeUniform(renderer);
	}
	return renderer;
}
